import asyncio
from typing import TYPE_CHECKING, List, Literal

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Bot, Message, MessageEvent, MessageSegment
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from ..config import config
from ..data.gamekee import game_kee_calender
from ..data.schaledb import schale_calender
from ..help import FT_E, FT_S
from ..util import send_forward_msg

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "日程表",
        "trigger_method": "指令",
        "trigger_condition": "ba日程表",
        "brief_des": "查看活动日程表",
        "detail_des": (
            "查看当前未结束的卡池、活动以及起止时间\n"
            "默认展示来自GameKee源的所有服务器日程\n"
            "可以使用下方的指令参数指定数据源和展示的服务器\n"
            " \n"
            "可以在指令后带参数，每个参数请使用空格分隔\n"
            "参数列表：\n"
            "- 使用SchaleDB数据源：\n"
            f"  {FT_S}夏莱{FT_E} / {FT_S}沙勒{FT_E} / {FT_S}s{FT_E} / {FT_S}schale{FT_E} / {FT_S}schaledb{FT_E}\n"
            "- 展示日服日程：\n"
            f"  {FT_S}日{FT_E} / {FT_S}日服{FT_E} / {FT_S}j{FT_E} / {FT_S}jp{FT_E} / {FT_S}japan{FT_E}\n"
            "- 展示国际服日程：\n"
            f"  {FT_S}国际{FT_E} / {FT_S}国际服{FT_E} / {FT_S}g{FT_E} / {FT_S}gl{FT_E} / {FT_S}global{FT_E}\n"
            "- 展示国服日程：\n"
            f"  {FT_S}国服{FT_E} / {FT_S}c{FT_E} / {FT_S}cn{FT_E} / {FT_S}china{FT_E} / {FT_S}chinese{FT_E}\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba日程表{FT_E} （GameKee源）\n"
            f"- {FT_S}ba日程表 schale{FT_E} （SchaleDB源，所有服务器）\n"
            f"- {FT_S}ba日程表 schale 日服 国际服{FT_E} （SchaleDB源，日服和国际服）"
        ),
    },
]


cmd_calender = on_command("ba日程表")


@cmd_calender.handle()
async def _(
    bot: Bot,
    event: MessageEvent,
    matcher: Matcher,
    cmd_arg: Message = CommandArg(),
):
    args: List[str] = cmd_arg.extract_plain_text().strip().lower().split()

    gamekee = True
    servers: List[Literal["Jp", "Global", "Cn"]] = []

    if any((x in ("夏莱", "沙勒", "s", "schale", "schaledb")) for x in args):
        gamekee = False
    if any((x in ("日", "日服", "j", "jp", "japan")) for x in args):
        servers.append("Jp")
    if any((x in ("国际", "国际服", "g", "gl", "global")) for x in args):
        servers.append("Global")
    if any((x in ("国服", "c", "cn", "china", "chinese")) for x in args):
        servers.append("Cn")

    if not servers:
        servers = ["Jp", "Global", "Cn"]

    if gamekee:
        task = game_kee_calender(servers)
    else:
        task = asyncio.gather(*(schale_calender(x) for x in servers))

    await matcher.send("正在绘制图片，请稍等")
    try:
        pics = await task
    except Exception:
        logger.exception("绘制日程表图片出错")
        await matcher.finish("绘制日程表图片出错，请检查后台输出")

    if not pics:
        await matcher.finish("没有获取到日程表数据")

    messages = [MessageSegment.image(x) for x in pics]

    if config.ba_use_forward_msg:
        await send_forward_msg(bot, event, messages)
        return

    await matcher.finish(Message(messages))



import json
from io import BytesIO
from pathlib import Path
from typing import Union, cast

from nonebot import get_available_plugin_names, logger, require
from nonebot.adapters.onebot.v11 import MessageSegment
from nonebot.internal.adapter import Message
from nonebot.internal.matcher import Matcher
from nonebot.params import CommandArg
from PIL import Image
from pil_utils.fonts import get_proper_font

from ..command import help_list
from .const import FT_E, FT_S

# import importlib.util
# if importlib.util.find_spec("nonebot_plugin_PicMenu") is None:

if "nonebot_plugin_PicMenu" not in get_available_plugin_names():
    raise ImportError

require("nonebot_plugin_PicMenu")

from nonebot_plugin_PicMenu import menu_manager  # noqa: E402

usage = f"请使用指令 [ba帮助 {FT_S}功能名称或序号{FT_E}] 查看某功能详细介绍"
extra = {"menu_template": "default", "menu_data": help_list}


def save_img_to_io(img: Image.Image):
    img = img.convert("RGB")
    img_io = BytesIO()
    img.save(img_io, "jpeg")
    img_io.seek(0)
    return img_io


async def help_handle(matcher: Matcher, arg_msg: Message = CommandArg()):
    arg = arg_msg.extract_plain_text().strip()
    img = cast(
        Union[str, Image.Image],
        (
            menu_manager.generate_plugin_menu_image("BAWiki")
            if not arg
            else menu_manager.generate_func_details_image("BAWiki", arg)
        ),
    )

    if isinstance(img, str):
        await matcher.finish(f'出错了，可能未找到功能 "{arg}"')

    await matcher.finish(MessageSegment.image(save_img_to_io(img)))


# 给 PicMenu 用户上个默认字体
def set_pic_menu_font():
    pic_menu_dir = Path.cwd() / "menu_config"
    pic_menu_config = pic_menu_dir / "config.json"

    if not pic_menu_dir.exists():
        pic_menu_dir.mkdir(parents=True)

    if (not pic_menu_config.exists()) or (
        json.loads(pic_menu_config.read_text(encoding="u8")).get("default")
        == "font_path"
    ):
        path = str(get_proper_font("国").path.resolve())
        pic_menu_config.write_text(json.dumps({"default": path}), encoding="u8")
        logger.info("检测到 PicMenu 已加载并且未配置字体，已自动帮您配置系统字体")


try:
    set_pic_menu_font()
except Exception:
    logger.exception("配置 PicMenu 字体失败")



import datetime
from typing import TYPE_CHECKING, Literal, Optional

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Message
from nonebot.matcher import Matcher
from nonebot.params import CommandArg
from nonebot.typing import T_State

from ..data.bawiki import db_future
from ..help import FT_E, FT_S

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "更新前瞻",
        "trigger_method": "指令",
        "trigger_condition": "ba千里眼",
        "brief_des": "查询国际服/国服未来的卡池与活动",
        "detail_des": (
            "发送当前或指定日期的国际服/国服未来卡池与活动列表\n"
            "图片作者 B站@夜猫咪喵喵猫\n"
            " \n"
            "参数可以指定起始日期以及列表个数，但同时指定时请将日期放在列表个数前面，以空格分隔\n"
            "参数中含有 `全` 或 `a` 时会发送整张前瞻图\n"
            " \n"
            "参数以 `国际服` 或 `国际` 或 `global` 或 `g` 开头时会发送 国际服 前瞻图\n"
            "参数以 `国服` 或 `国` 或 `chinese` 或 `c` 开头时会发送 国服 前瞻图\n"
            "如果参数不以这些词开头时，默认发送国际服前瞻图\n"
            " \n"
            "日期格式可以为（Y代表4位数年，m代表月，d代表日）：\n"
            f"- {FT_S}Y/m/d{FT_E}；{FT_S}m/d{FT_E}\n"
            f"- {FT_S}Y-m-d{FT_E}；{FT_S}m-d{FT_E}\n"
            f"- {FT_S}Y年m月d日{FT_E}；{FT_S}m月d日{FT_E}\n"
            " \n"
            "可以用这些指令触发：\n"
            f"- {FT_S}ba千里眼{FT_E}\n"
            f"- {FT_S}ba前瞻{FT_E}\n"
            " \n"
            "有以下指令别名：\n"
            f"- {FT_S}ba国服千里眼{FT_E} 或 {FT_S}ba国服前瞻{FT_E} -> {FT_S}ba千里眼 国服{FT_E}\n"
            f"- {FT_S}ba国际服千里眼{FT_E} 或 {FT_S}ba国际服前瞻{FT_E} -> {FT_S}ba千里眼 国际服{FT_E}\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba千里眼{FT_E}\n"
            f"- {FT_S}ba千里眼 all{FT_E}\n"
            f"- {FT_S}ba千里眼 3{FT_E}\n"
            f"- {FT_S}ba千里眼 11/15{FT_E}\n"
            f"- {FT_S}ba千里眼 11/15 3{FT_E}\n"
            f"- {FT_S}ba千里眼 国服{FT_E}\n"
            f"- {FT_S}ba千里眼 国际服 11/15{FT_E}"
        ),
    },
]


GLOBAL_PFX = ("国际服", "国际", "global", "g")
CHINESE_PFX = ("国服", "国", "chinese", "c")

cmd_future = on_command("ba千里眼", aliases={"ba前瞻"})
cmd_global_future = on_command(
    "ba国际服千里眼",
    aliases={"ba国际服前瞻"},
    state={"future_type": "global"},
)
cmd_chinese_future = on_command(
    "ba国服千里眼",
    aliases={"ba国服前瞻"},
    state={"future_type": "chinese"},
)


@cmd_future.handle()
@cmd_global_future.handle()
@cmd_chinese_future.handle()
async def _(matcher: Matcher, state: T_State, arg: Message = CommandArg()):
    args = arg.extract_plain_text().lower().strip()

    future_type: Optional[Literal["global", "chinese"]] = state.get("future_type")
    if not future_type:
        used_global_pfx = next(filter(args.startswith, GLOBAL_PFX), None)
        used_chinese_pfx = next(filter(args.startswith, CHINESE_PFX), None)

        used_pfx = next((x for x in (used_chinese_pfx, used_global_pfx) if x), None)
        if used_pfx:
            args = args[len(used_pfx) :].strip()

        is_chinese = used_chinese_pfx
        future_type = "chinese" if is_chinese else "global"

    if "全" in args or "a" in args:
        await matcher.finish(await db_future(future_type, all_img=True))

    args = args.split()
    num = 3
    date = None
    if (args_len := len(args)) == 1:
        if args[0].isdigit():
            num = args[0]
        else:
            date = args[0]
    elif args_len > 1:
        date = args[0].strip()
        num = args[-1].strip()

    if date:
        parsed_date = None
        for f in ["%Y/%m/%d", "%Y-%m-%d", "%Y年%m月%d日", "%m/%d", "%m-%d", "%m月%d日"]:
            try:
                parsed_date = datetime.datetime.strptime(date.replace(" ", ""), f)
                break
            except ValueError:
                pass
        if not parsed_date:
            await matcher.finish("日期格式不正确！")
        date = parsed_date
        if date.year == 1900:
            now = datetime.datetime.now().replace(
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            )
            date = date.replace(year=now.year)
            if date < now:
                date = date.replace(year=now.year + 1)

    if isinstance(num, str) and ((not num.isdigit()) or (num := int(num)) < 1):
        await matcher.finish("前瞻项目数量格式不正确！")

    await matcher.finish(await db_future(future_type, date or None, num))



from typing import TYPE_CHECKING

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Message
from nonebot.log import logger
from nonebot.matcher import Matcher

from ..data.bawiki import db_wiki_craft
from ..help import FT_E, FT_S

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "制造一图流",
        "trigger_method": "指令",
        "trigger_condition": "ba制造",
        "brief_des": "查询制造功能机制图",
        "detail_des": (
            "发送游戏内制造功能的一图流介绍\n"
            "图片作者 B站@夜猫咪喵喵猫\n"
            " \n"
            "可以用这些指令触发：\n"
            f"- {FT_S}ba制造{FT_E}\n"
            f"- {FT_S}ba制作{FT_E}\n"
            f"- {FT_S}ba合成{FT_E}"
        ),
    },
]


cmd_craft_wiki = on_command("ba制造", aliases={"ba合成", "ba制作"})


@cmd_craft_wiki.handle()
async def _(matcher: Matcher):
    try:
        im = await db_wiki_craft()
    except Exception:
        logger.exception("获取合成wiki图片错误")
        await matcher.finish("获取图片失败，请检查后台输出")

    await matcher.finish(Message(im))



import asyncio
from typing import TYPE_CHECKING

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Message
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from ..data.bawiki import db_get_event_alias, db_wiki_event
from ..data.schaledb import find_current_event, schale_get_config
from ..help import FT_E, FT_S
from ..util import recover_alia, splice_msg

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "活动一图流",
        "trigger_method": "指令",
        "trigger_condition": "ba活动",
        "brief_des": "查询活动攻略图",
        "detail_des": (
            "发送当前或指定活动一图流攻略图，可能会附带活动特殊机制等\n"
            "图片作者 B站@夜猫咪喵喵猫\n"
            " \n"
            "指令默认发送日服和国际服当前的活动攻略\n"
            "指令后面跟`日`或`j`开头的文本代表查询日服当前活动攻略，带以`国际`或`g`、`国`或`c`开头的文本同理\n"
            "跟其他文本则代表指定活动名称\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba活动{FT_E}\n"
            f"- {FT_S}ba活动 日{FT_E}\n"
            f"- {FT_S}ba活动 温泉浴场{FT_E}"
        ),
    },
]


cmd_event_wiki = on_command("ba活动")


@cmd_event_wiki.handle()
async def _(matcher: Matcher, cmd_arg: Message = CommandArg()):
    arg = cmd_arg.extract_plain_text().lower().strip()

    keys = {
        0: ("日", "j"),
        1: ("国际", "g"),
        2: ("国", "c"),
    }

    server = []
    for k, v in keys.items():
        if (not arg) or arg.startswith(v):
            server.append(k)
            for kw in v:
                arg = arg.replace(kw, "", 1)

    events = []
    if server:
        try:
            common = await schale_get_config()
            for s in server:
                ev = common["Regions"][s]["CurrentEvents"]
                if e := find_current_event(ev):
                    events.append((e[0]["event"]) % 10000)
        except Exception:
            logger.exception("获取当前活动失败")
            await matcher.finish("获取当前活动失败")

        if not events:
            await matcher.finish("当前服务器没有正在进行的活动")

    else:
        events.append(recover_alia(arg, await db_get_event_alias()))

    try:
        ret = await asyncio.gather(*[db_wiki_event(x) for x in events])
    except Exception:
        logger.exception("获取活动wiki出错")
        await matcher.finish("获取图片出错，请检查后台输出")

    await matcher.finish(splice_msg(ret))



import asyncio
from typing import TYPE_CHECKING

from nonebot import on_command
from nonebot.adapters.onebot.v11 import ActionFailed, Message, MessageSegment
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from ..config import config
from ..data.bawiki import db_get_extra_l2d_list, recover_stu_alia, schale_to_gamekee
from ..data.gamekee import game_kee_get_stu_cid_li, game_kee_grab_l2d
from ..data.schaledb import draw_fav_li, get_fav_li, schale_get_stu_dict
from ..help import FT_E, FT_S
from ..util import RespType as Rt
from ..util import async_req

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "羁绊查询",
        "trigger_method": "指令",
        "trigger_condition": "ba羁绊",
        "brief_des": "查询学生解锁L2D需求的羁绊等级",
        "detail_des": (
            "使用学生名称查询该学生解锁L2D看板需求的羁绊等级以及L2D预览，"
            "或者使用羁绊等级级数查询哪些学生达到该等级时解锁L2D看板\n"
            "使用学生名称查询时支持部分学生别名\n"
            " \n"
            "可以用这些指令触发：\n"
            f"- {FT_S}ba羁绊{FT_E}\n"
            f"- {FT_S}ba好感度{FT_E}\n"
            f"- {FT_S}bal2d{FT_E}\n"
            f"- {FT_S}balive2d{FT_E}\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba羁绊 xcw{FT_E}\n"
            f"- {FT_S}ba羁绊 9{FT_E}"
        ),
    },
]


cmd_fav = on_command(
    "ba好感度",
    aliases={"ba羁绊", "bal2d", "baL2D", "balive2d", "baLive2D"},
)


@cmd_fav.handle()
async def _(matcher: Matcher, cmd_arg: Message = CommandArg()):
    async def check_size(url: str) -> bool:
        headers = await async_req(url, method="HEAD", resp_type=Rt.HEADERS)
        size = headers.get("Content-Length", 0)
        if not size:
            logger.debug(f"HEAD {url} resp header has no Content-Length")
            return True
        ok = int(size) < 1024 * 1024 * 10
        if not ok:
            logger.warning(f"{url} size too large ({size * 1024:.2f} KB > 10240 KB)")
        return ok

    async def get_l2d(stu_name):
        if r := (await db_get_extra_l2d_list()).get(stu_name):
            return [f"{config.ba_bawiki_db_url}{x}" for x in r]

        cid = (await game_kee_get_stu_cid_li()).get(stu_name)
        if cid:
            l2d_list = await game_kee_grab_l2d(cid)
            checked_resp = await asyncio.gather(
                *(check_size(x) for x in l2d_list),
            )
            return [x for x, y in zip(l2d_list, checked_resp) if y]

        return None

    arg = cmd_arg.extract_plain_text().strip()
    if not arg:
        await matcher.finish("请提供学生名称或所需的羁绊等级")

    # 好感度等级
    if arg.isdigit():
        arg = int(arg)

        try:
            li = await get_fav_li(arg)
        except Exception:
            logger.exception("获取 SchaleDB 学生数据失败")
            await matcher.finish("获取 SchaleDB 学生数据失败，请检查后台输出")

        if not li:
            await matcher.finish(f"没有学生在羁绊等级{arg}时解锁L2D")

        try:
            p = await draw_fav_li(li)
        except Exception:
            logger.exception("绘制图片出错")
            await matcher.finish("绘制图片出错，请检查后台输出")

        await matcher.finish(f"羁绊等级 {arg} 时解锁L2D的学生有以下这些：" + MessageSegment.image(p))

    # 学生名称
    arg = await recover_stu_alia(arg)

    try:
        ret = await schale_get_stu_dict()
    except Exception:
        logger.exception("获取学生列表出错")
        await matcher.finish("获取学生列表表出错，请检查后台输出")

    if stu := ret.get(arg):
        if not (lvl := stu["MemoryLobby"]):
            await matcher.finish("该学生没有L2D")

        try:
            pics = await get_l2d(await schale_to_gamekee(arg))
        except Exception:
            logger.exception("下载 L2D 图片出错")
            await matcher.finish("获取 L2D 图片列表出错，请检查后台输出")

        im = Message() + f'{stu["Name"]} 在羁绊等级 {lvl[0]} 时即可解锁L2D\n'
        if pics:
            try:
                images = await asyncio.gather(
                    *(async_req(x, resp_type=Rt.BYTES) for x in pics),
                )
            except Exception:
                logger.exception("下载 L2D 图片出错")
                await matcher.finish("下载 L2D 图片出错，请检查后台输出")
            image_seg = "L2D预览：" + Message(MessageSegment.image(x) for x in images)

        else:
            im += (
                "没找到该学生的L2D看板\n"
                "可能原因：\n"
                "- GameKee页面爬取不到角色L2D图片\n"
                "- GameKee和插件没有收录该学生的L2D\n"
            )
            await matcher.finish(im)

        try:
            await matcher.finish(im + image_seg)
        except ActionFailed as e:
            if image_seg:
                logger.warning(f"Failed to send message: ActionFailed: {e}")
                await matcher.finish(im + "抱歉，L2D 图片被风控了，或许是因为太涩了……")
            raise

    await matcher.finish("未找到学生")



from ..resource import BA_LOGO_JS_PATH
from .playwright import get_routed_page


async def get_logo(text_l: str, text_r: str, transparent_bg: bool = True) -> str:
    async with get_routed_page() as page:
        return await page.evaluate(
            BA_LOGO_JS_PATH.read_text(encoding="u8"),
            [text_l, text_r, transparent_bg],
        )



from typing import TYPE_CHECKING

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Message, MessageSegment
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from ..data.gamekee import extract_content_pic, get_level_list
from ..help import FT_E, FT_S
from ..util import RespType as Rt
from ..util import async_req

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "关卡攻略",
        "trigger_method": "指令",
        "trigger_condition": "ba关卡",
        "brief_des": "获取关卡攻略",
        "detail_des": (
            "获取指定关卡攻略\n"
            "来源：GameKee\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba关卡 1-1{FT_E}\n"
            f"- {FT_S}ba关卡 H1-1{FT_E}"
        ),
    },
]


cmd_level_guide = on_command("ba关卡")


@cmd_level_guide.handle()
async def _(matcher: Matcher, arg: Message = CommandArg()):
    arg_str = arg.extract_plain_text().strip().upper()
    if not arg_str:
        await matcher.finish("请输入关卡名称")

    try:
        levels = await get_level_list()
    except Exception:
        logger.exception("获取关卡列表失败")
        await matcher.finish("获取关卡列表失败，请检查后台输出")

    if arg_str not in levels:
        await matcher.finish("未找到该关卡，请检查关卡名称是否正确")

    cid = levels[arg_str]
    try:
        imgs = await extract_content_pic(cid)
    except Exception:
        logger.exception("获取攻略图片失败")
        await matcher.finish("获取攻略图片失败，请检查后台输出")

    msg = Message()
    msg += f"https://ba.gamekee.com/{cid}.html\n"
    msg += [MessageSegment.image(await async_req(x, resp_type=Rt.BYTES)) for x in imgs]
    await matcher.finish(msg)



import asyncio
import shutil
from base64 import b64encode
from datetime import datetime
from enum import Enum
from io import BytesIO
from typing import (
    Any,
    AsyncIterable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)
from typing_extensions import Unpack
from urllib.parse import urljoin

import anyio
import jinja2
import matplotlib.dates as mdates
import matplotlib.ticker as mticker
import pytz
from matplotlib import pyplot
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from nonebot import logger
from nonebot_plugin_htmlrender import get_new_page
from playwright.async_api import Route, ViewportSize
from pydantic import BaseModel, Field, parse_obj_as, validator
from yarl import URL

from ..config import config
from ..resource import (
    CACHE_DIR,
    RES_SHITTIM_TEMPLATES_DIR,
    SHITTIM_UTIL_CSS_PATH,
    SHITTIM_UTIL_JS_PATH,
)
from ..util import (
    AsyncReqKwargs,
    RespType,
    base_async_req,
    camel_case,
    wrapped_alru_cache,
)
from .playwright import RES_ROUTE_URL, bawiki_router, get_template_renderer

if not config.ba_shittim_key:
    logger.warning("API Key 未配置，关于什亭之匣的功能将会不可用！")
    logger.warning("请访问 https://arona.icu/about 查看获取 API Key 的方式！")


SHITTIM_CACHE_DIR = CACHE_DIR / "shittim"
if config.ba_auto_clear_cache_path and SHITTIM_CACHE_DIR.exists():
    shutil.rmtree(SHITTIM_CACHE_DIR)
if not SHITTIM_CACHE_DIR.exists():
    SHITTIM_CACHE_DIR.mkdir(parents=True)


T = TypeVar("T")

SERVER_NAME_MAP = {
    1: "官服",
    2: "B服",
}
RANK_DATA_TYPE_NAME_MAP = {
    1: "排名",
    2: "档线",
}
HARD_FULLNAME_MAP = {
    "EX": "Extreme",
    "HC": "Hardcore",
    "VH": "VeryHard",
    "H": "Hard",
    "N": "Normal",
}
RAID_ANALYSIS_URL = urljoin(config.ba_shittim_url, "raidAnalyse")
TIMEZONE_SHANGHAI = pytz.timezone("Asia/Shanghai")

async_req = wrapped_alru_cache(ttl=config.ba_shittim_req_cache_ttl, maxsize=None)(
    base_async_req,
)
template_env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(RES_SHITTIM_TEMPLATES_DIR),
    enable_async=True,
)


# region Pagination


class PaginationCallable(Protocol, Generic[T]):
    # (resp, is_last_page)
    async def __call__(
        self,
        page: int,
        size: int,
        delay: float,
    ) -> Tuple[Optional[List[T]], bool]:
        ...


class IterPFKwargs(TypedDict, total=False):
    page: int
    size: int
    delay: float


def iter_pagination_func(**kwargs: Unpack[IterPFKwargs]):
    page = kwargs.get("page", 1)
    size = kwargs.get("size", 100)
    delay = kwargs.get("delay", 0.0)

    def decorator(func: PaginationCallable[T]) -> Callable[[], AsyncIterable[T]]:
        async def wrapper():
            while True:
                resp, last_page = await func(page, size, delay)
                if resp:
                    for x in resp:
                        yield x
                if last_page:
                    break

        return wrapper

    return decorator


async def async_iter_all(iterator: AsyncIterable[T]) -> List[T]:
    return [x async for x in iterator]


# endregion


# region enums


class ServerType(Enum):
    Official = 1
    Bilibili = 2


class RankDataType(Enum):
    Rank = 1
    Score = 2


# endregion


# region models


def validator_time(v: str):
    try:
        return (
            datetime.strptime(v, "%Y-%m-%d %H:%M")
            .replace(tzinfo=TIMEZONE_SHANGHAI)
            .astimezone()
        )
    except ValueError as e:
        raise ValueError(f"Time `{v}` format error") from e


def validator_time_as_local(v: datetime) -> datetime:
    return v.astimezone()


class CamelAliasModel(BaseModel):
    class Config:
        alias_generator = camel_case


class PaginationModel(CamelAliasModel):
    page: int
    size: int
    total_pages: int
    last_page: bool


class SeasonMap(CamelAliasModel):
    key: str
    value: str


class Season(CamelAliasModel):
    season: int
    season_map: SeasonMap = Field(alias="map")
    boss_id: int
    boss: str
    start_time: datetime
    end_time: datetime

    _validator_time = validator(
        "start_time",
        "end_time",
        pre=True,
        allow_reuse=True,
    )(validator_time)


class Character(CamelAliasModel):
    has_weapon: bool
    is_assist: bool
    level: int
    slot_index: int
    star_grade: int
    unique_id: int
    bullet_type: str
    tactic_role: str


class TryNumberInfo(CamelAliasModel):
    try_number: int
    main_characters: List[Character]
    support_characters: List[Character]


class RankSummary(CamelAliasModel):
    rank: int
    best_ranking_point: int
    hard: str
    battle_time: str

    @property
    def hard_fullname(self) -> str:
        return HARD_FULLNAME_MAP.get(self.hard, self.hard)


class RankRecord(RankSummary):
    level: int
    nickname: str
    represent_character_unique_id: int
    tier: int
    boss_id: int
    try_number_infos: Optional[List[TryNumberInfo]]
    record_time: datetime

    _validator_time = validator(
        "record_time",
        allow_reuse=True,
    )(validator_time_as_local)


class Rank(PaginationModel):
    records: List[RankRecord]


class RaidChart(CamelAliasModel):
    data: Dict[int, List[int]]
    time: List[datetime]

    _validator_time = validator(
        "time",
        each_item=True,
        allow_reuse=True,
    )(validator_time_as_local)


class ParticipationChart(CamelAliasModel):
    key: List[datetime]
    value: List[int]

    _validator_time = validator(
        "key",
        each_item=True,
        allow_reuse=True,
    )(validator_time_as_local)


# endregion


# region api


request_lock = asyncio.Lock()


async def shittim_get(url: str, **kwargs: Unpack[AsyncReqKwargs]) -> Any:
    if not config.ba_shittim_key:
        raise ValueError("`BA_SHITTIM_KEY` not set")

    kwargs["base_urls"] = config.ba_shittim_api_url
    kwargs["proxies"] = config.ba_shittim_proxy

    headers = kwargs.get("headers") or {}
    headers["Authorization"] = f"ba-token {config.ba_shittim_key}"
    kwargs["headers"] = headers

    limit_qps = config.ba_shittim_request_delay > 0
    if limit_qps:
        kwargs["sleep"] = config.ba_shittim_request_delay
        await request_lock.acquire()

    try:
        resp = await async_req(url, **kwargs)
    finally:
        if limit_qps and request_lock.locked():
            request_lock.release()

    if (code := resp.get("code")) != 200:
        params = kwargs.get("params")
        logger.warning(
            f"Shittim API `{url}` returned error code {code}, {params=}, {resp=}",
        )

    return resp["data"]


async def get_season_list() -> List[Season]:
    return parse_obj_as(List[Season], await shittim_get("api/season/list"))


def get_rank_list(
    server: ServerType,
    data_type: RankDataType,
    season: int,
    **pf_kwargs: Unpack[IterPFKwargs],
) -> AsyncIterable[RankRecord]:
    @iter_pagination_func(**pf_kwargs)
    async def iterator(page: int, size: int, delay: float):
        ret = parse_obj_as(
            Rank,
            await shittim_get(
                f"api/rank/list/{server.value}/{data_type.value}/{season}",
                params={"page": page, "size": size},
                sleep=delay,
            ),
        )
        return ret.records, True if (not ret.records) else ret.last_page

    return iterator()


async def get_rank_list_top(
    server: ServerType,
    season: int,
) -> List[RankSummary]:
    return parse_obj_as(
        List[RankSummary],
        await shittim_get(
            "api/rank/list_top",
            params={"server": server.value, "season": season},
        ),
    )


async def get_rank_list_by_last_rank(
    server: ServerType,
    season: int,
) -> List[RankSummary]:
    return parse_obj_as(
        List[RankSummary],
        await shittim_get(
            "api/rank/list_by_last_rank",
            params={"server": server.value, "season": season},
        ),
    )


async def get_raid_chart_data(server: ServerType, season: int) -> RaidChart:
    return parse_obj_as(
        RaidChart,
        await shittim_get(f"raid/new/charts/{server.value}", params={"s": season}),
    )


async def get_participation_chart_data(
    server: ServerType,
    season: int,
) -> ParticipationChart:
    return parse_obj_as(
        ParticipationChart,
        await shittim_get(
            "api/rank/season/lastRank/charts",
            params={"server": server.value, "season": season},
        ),
    )


async def get_alice_friends(server: ServerType) -> Dict[int, RankRecord]:
    return parse_obj_as(
        Dict[int, RankRecord],
        await shittim_get("api/rank/list_20001", params={"server": server.value}),
    )


async def get_diligent_achievers(server: ServerType) -> Dict[int, RankRecord]:
    return parse_obj_as(
        Dict[int, RankRecord],
        await shittim_get("api/rank/list_1", params={"server": server.value}),
    )


async def get_student_icon(student_id: Union[int, str]) -> bytes:
    filename = f"{student_id}.png"
    path = anyio.Path(SHITTIM_CACHE_DIR / filename)

    if await path.exists():
        return await path.read_bytes()

    resp = await async_req(
        f"web_students_original_icon/{filename}",
        base_urls=config.ba_shittim_data_url,
        resp_type=RespType.BYTES,
        headers={"Referer": config.ba_shittim_api_url},
    )
    await path.write_bytes(resp)
    return resp


# endregion


# region render

VIEWPORT_SIZE = ViewportSize(width=880, height=1080)

CHART_SHOW_RANKS = [1, 1000, 2000, 4000, 8000, 20000]
CHART_W = 760
CHART_H = 480
DATE_FORMAT = "%m-%d %H:%M"
NUM_FORMAT = "{x:,.0f}"
DATE_FORMATTER = mdates.DateFormatter(DATE_FORMAT)
NUM_FORMATTER = mticker.StrMethodFormatter(NUM_FORMAT)


def get_figure() -> "Figure":
    figure = pyplot.figure()
    figure.set_size_inches(CHART_W / figure.dpi, CHART_H / figure.dpi)
    return figure


def save_figure(figure: Figure) -> bytes:
    bio = BytesIO()
    figure.savefig(bio, transparent=True, format="png")
    return bio.getvalue()


def ax_settings(ax: Axes) -> None:
    ax.grid()
    ax.legend(loc="lower right")
    ax.xaxis.set_major_formatter(DATE_FORMATTER)
    ax.yaxis.set_major_formatter(NUM_FORMATTER)
    ax.tick_params(axis="x", labelrotation=15)


def render_raid_chart(data: RaidChart) -> bytes:
    figure = get_figure()

    ax = figure.add_subplot()
    for key in CHART_SHOW_RANKS:
        if (key not in data.data) or (not (y := data.data[key])):
            continue
        x = data.time[: len(y)]
        ax.plot(
            x,
            y,
            label=(
                f"{' ' * (10 - (len(str(key)) * 2))}{key} | "
                f"{x[-1].strftime(DATE_FORMAT)} | "
                f"{NUM_FORMAT.format(x=y[-1])}"
            ),
        )
    ax_settings(ax)

    figure.tight_layout()
    return save_figure(figure)


def render_participation_chart(data: ParticipationChart) -> bytes:
    figure = get_figure()

    ax = figure.add_subplot()
    ax.plot(
        data.key,
        data.value,
        label=(
            "Participants | "
            f"{data.key[-1].strftime(DATE_FORMAT)} | "
            f"{NUM_FORMAT.format(x=data.value[-1])}"
        ),
    )
    ax_settings(ax)

    figure.tight_layout()
    return save_figure(figure)


def to_b64_url(data: bytes) -> str:
    return f"data:image/png;base64,{b64encode(data).decode()}"


async def render_raid_rank(
    server_name: str,
    data_type_name: str,
    season: Season,
    rank_list_top: List[RankSummary],
    rank_list_by_last_rank: List[RankSummary],
    rank_list: List[RankRecord],
    raid_chart: RaidChart,
    participation_chart: ParticipationChart,
) -> bytes:
    template = template_env.get_template("content_raid_rank.html.jinja")
    raid_chart_url = to_b64_url(render_raid_chart(raid_chart))
    participation_chart_url = to_b64_url(
        render_participation_chart(participation_chart),
    )
    return await get_template_renderer(
        template,
        selector=".wrapper",
        viewport=VIEWPORT_SIZE,
    )(
        server_name=server_name,
        data_type_name=data_type_name,
        season=season,
        rank_list_top=rank_list_top,
        rank_list_by_last_rank=rank_list_by_last_rank,
        rank_list=rank_list,
        raid_chart_url=raid_chart_url,
        participation_chart_url=participation_chart_url,
        shittim_url=config.ba_shittim_url,
    )


async def render_rank_detail(
    title: str,
    season_list: List[Season],
    rank_list: Dict[int, RankRecord],
) -> bytes:
    template = template_env.get_template("content_rank_detail.html.jinja")
    return await get_template_renderer(
        template,
        selector=".wrapper",
        viewport=VIEWPORT_SIZE,
    )(
        title=title,
        season_list=season_list,
        rank_list=sorted(rank_list.items(), key=lambda x: x[0], reverse=True),
        shittim_url=config.ba_shittim_url,
    )


async def render_raid_analysis() -> bytes:
    async with get_new_page(viewport=VIEWPORT_SIZE) as page:
        await page.goto(RAID_ANALYSIS_URL, wait_until="networkidle")

        await page.evaluate(SHITTIM_UTIL_JS_PATH.read_text(encoding="u8"))
        await page.add_style_tag(content=SHITTIM_UTIL_CSS_PATH.read_text(encoding="u8"))

        elem = await page.query_selector(".content")
        assert elem
        return await elem.screenshot(type="jpeg")


# endregion


RES_TYPE_SHITTIM_STUDENT_ICON = "shittim_student_icon"


@bawiki_router(rf"^{RES_ROUTE_URL}/{RES_TYPE_SHITTIM_STUDENT_ICON}/(\d+)$")
async def _(url: URL, route: Route, **_):
    student_id = url.parts[-1]
    icon = await get_student_icon(student_id)
    return await route.fulfill(body=icon, content_type="image/png")



from typing import Optional

from nonebot import get_driver
from pydantic import BaseModel, Field, HttpUrl


class Cfg(BaseModel):
    ba_proxy: Optional[str] = None
    ba_shittim_proxy: Optional[str] = None
    ba_gacha_cool_down: int = 0
    ba_voice_use_card: bool = False
    ba_use_forward_msg: bool = True
    ba_screenshot_timeout: int = 60
    ba_disable_classic_gacha: bool = False
    ba_gacha_max: int = 200
    ba_illegal_limit: int = 3
    ba_arona_set_alias_only_su: bool = False

    ba_gamekee_url: HttpUrl = Field("https://ba.gamekee.com/")
    ba_schale_url: HttpUrl = Field("https://schale.gg/")
    ba_bawiki_db_url: HttpUrl = Field("https://bawiki.lgc2333.top/")
    ba_arona_api_url: HttpUrl = Field("https://arona.diyigemt.com/")
    ba_arona_cdn_url: HttpUrl = Field("https://arona.cdn.diyigemt.com/")
    ba_shittim_url: HttpUrl = Field("https://arona.icu/")
    ba_shittim_api_url: HttpUrl = Field("https://api.arona.icu/")
    ba_shittim_data_url: HttpUrl = Field("https://data.ba.benx1n.com/")

    ba_shittim_key: Optional[str] = None
    ba_shittim_request_delay: float = 0

    ba_req_retry: int = 1
    ba_req_cache_ttl: int = 10800  # 3 hrs
    ba_shittim_req_cache_ttl: int = 600  # 10 mins
    ba_req_timeout: Optional[float] = 10.0
    ba_auto_clear_cache_path: bool = False


config = Cfg.parse_obj(get_driver().config.dict())



import mimetypes
import re
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
)

import anyio
import jinja2
from nonebot import logger
from nonebot_plugin_htmlrender import get_new_page
from pil_utils.fonts import Font, get_proper_font
from playwright.async_api import Page, Request, Route
from yarl import URL

from ..resource import EMPTY_HTML_PATH, RES_DIR

PWRouter = Callable[[Route, Request], Awaitable[None]]
BAWikiRouterFunc = Callable[..., Awaitable[None]]
TRF = TypeVar("TRF", bound=BAWikiRouterFunc)

RES_ROUTE_URL = "https://bawiki.res"
RES_TYPE_FONT = "font"


registered_routers: List["BAWikiRouter"] = []


@dataclass
class BAWikiRouter:
    pattern: Union[str, re.Pattern]
    func: BAWikiRouterFunc
    priority: int


def bawiki_router(
    pattern: Union[str, re.Pattern],
    flags: Optional[re.RegexFlag] = None,
    priority: int = 0,
):
    if not isinstance(pattern, re.Pattern):
        pattern = re.compile(pattern, flags=flags or 0)

    def wrapper(func: TRF) -> TRF:
        registered_routers.append(BAWikiRouter(pattern, func, priority))
        # 低 priority 的 BAWikiRouter 应最先运行，
        # 因为 playwright 后 route 的先运行，所以要反过来排序
        registered_routers.sort(key=lambda r: r.priority, reverse=True)
        logger.debug(f"Registered router: {pattern=}, {priority=}")
        return func

    return wrapper


async def route_page(page: Page, router: BAWikiRouter):
    async def wrapped(route: Route, request: Request):
        url = URL(request.url)
        logger.debug(f"Requested routed URL: {url.human_repr()}")
        match = re.search(router.pattern, request.url)
        assert match
        return await router.func(match=match, url=url, route=route, request=request)

    await page.route(router.pattern, wrapped)


@asynccontextmanager
async def get_routed_page(**kwargs) -> AsyncIterator[Page]:
    async with get_new_page(**kwargs) as page:
        for router in registered_routers:
            await route_page(page, router)
        await page.goto(RES_ROUTE_URL)
        yield page


async def render_html(
    html: str,
    selector: str = "body",
    img_format: Literal["png", "jpeg"] = "jpeg",
    **page_kwargs,
) -> bytes:
    from pathlib import Path

    Path("debug.html").write_text(html, encoding="u8")

    async with get_routed_page(**page_kwargs) as page:
        await page.set_content(html)
        elem = await page.query_selector(selector)
        assert elem
        return await elem.screenshot(type=img_format)


def get_template_renderer(
    template: jinja2.Template,
    selector: str = "body",
    img_format: Literal["png", "jpeg"] = "jpeg",
    **page_kwargs,
) -> Callable[..., Awaitable[bytes]]:
    async def renderer(**template_kwargs) -> bytes:
        html = await template.render_async(**template_kwargs)
        return await render_html(html, selector, img_format, **page_kwargs)

    return renderer


# region routers


@bawiki_router(rf"^{RES_ROUTE_URL}/?$")
async def _(route: Route, **_):
    html = EMPTY_HTML_PATH.read_text()
    return await route.fulfill(body=html, content_type="text/html")


@bawiki_router(rf"^{RES_ROUTE_URL}/{RES_TYPE_FONT}/([^/]+?)/?$")
async def _(url: URL, route: Route, **_):
    family = url.parts[-1]
    style = url.query.get("style", "normal")
    weight = url.query.get("weight", "normal")
    try:
        font = Font.find(
            family,
            style=style,
            weight=weight,
            fallback_to_default=False,
        )
    except Exception:
        logger.info(f"Font `{family}` not found, use fallback font")
        font = get_proper_font("国", style=style, weight=weight)  # type: ignore

    file_path = anyio.Path(font.path)
    mime = f"font/{file_path.suffix[1:]}"

    logger.debug(f"Resolved font `{family}`, file path: {file_path}")
    return await route.fulfill(body=await file_path.read_bytes(), content_type=mime)


@bawiki_router(rf"^{RES_ROUTE_URL}/(.+)$", priority=99)
async def _(url: URL, route: Route, **_):
    res_path = url.parts[1:]
    file_path = anyio.Path(RES_DIR.joinpath(*res_path))

    if not await file_path.exists():
        logger.debug(f"Resource `{res_path}` not found")
        return await route.abort()

    mime = mimetypes.guess_type(file_path)[0]
    logger.debug(
        f"Resolved resource `{'/'.join(res_path)}`, mimetype: {mime}, real path: {file_path}",
    )
    return await route.fulfill(
        body=await file_path.read_bytes(),
        content_type=mime or "application/octet-stream",
    )


# endregion



from pathlib import Path
from typing import TYPE_CHECKING

from nonebot import on_command
from nonebot.matcher import Matcher
from nonebot.permission import SUPERUSER

from ..help import FT_E, FT_S
from ..resource import CACHE_DIR
from ..util import clear_wrapped_alru_cache

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "清空缓存",
        "trigger_method": "超级用户 指令",
        "trigger_condition": "ba清空缓存",
        "brief_des": "清空插件请求缓存",
        "detail_des": (
            "手动清空插件请求网络缓存下来的数据，如API返回的数据\n"
            f"注：该指令只能由{FT_S}超级用户{FT_E}触发\n"
            " \n"
            "可以用这些指令触发：\n"
            f"- {FT_S}ba清空缓存{FT_E}\n"
            f"- {FT_S}ba清除缓存{FT_E}"
        ),
    },
]


def clear_cache_dir() -> int:
    counter = 0

    def run(path: Path):
        nonlocal counter
        for p in path.iterdir():
            if p.is_dir():
                run(p)
            else:
                p.unlink()
                counter += 1

    run(CACHE_DIR)
    return counter


cmd_clear_cache = on_command("ba清空缓存", aliases={"ba清除缓存"}, permission=SUPERUSER)


@cmd_clear_cache.handle()
async def _(matcher: Matcher):
    req_count = clear_wrapped_alru_cache()
    cache_count = clear_cache_dir()
    await matcher.finish(f"已清除 {req_count} 项请求缓存与 {cache_count} 项文件缓存～")



from typing import TYPE_CHECKING

from nonebot import logger, on_shell_command
from nonebot.adapters.onebot.v11 import MessageSegment
from nonebot.exception import ParserExit
from nonebot.matcher import Matcher
from nonebot.params import ShellCommandArgs
from nonebot.rule import ArgumentParser, Namespace

from ..data.logo_generate import get_logo
from ..help import FT_E, FT_S

if TYPE_CHECKING:
    from . import HelpList


help_list: "HelpList" = [
    {
        "func": "标题生成器",
        "trigger_method": "指令",
        "trigger_condition": "balogo",
        "brief_des": "生成 BA Logo 样式的图片",
        "detail_des": (
            "生成 BA Logo 样式的图片\n"
            "感谢 nulla2011/Bluearchive-logo 项目以 MIT 协议开源了图片绘制代码\n"
            " \n"
            "可以用这些指令触发：\n"
            f"- {FT_S}balogo{FT_E}\n"
            f"- {FT_S}baLogo{FT_E}\n"
            f"- {FT_S}baLOGO{FT_E}\n"
            f"- {FT_S}ba标题{FT_E}\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}balogo Blue Archive{FT_E}\n"
            f"- {FT_S}balogo -T Schale SenSei{FT_E}（使用参数 -T 加上背景）\n"
            f'- {FT_S}balogo "我是" "秦始皇"{FT_E}（包含空格的文本请使用引号包裹）'
        ),
    },
]


parser = ArgumentParser("balogo", add_help=False)
parser.add_argument("text_l")
parser.add_argument("text_r")
parser.add_argument("-T", "--no-transparent", action="store_true")

cmd_ba_logo = on_shell_command(
    "balogo",
    aliases={"baLogo", "baLOGO", "ba标题"},
    parser=parser,
)


@cmd_ba_logo.handle()
async def _(matcher: Matcher, err: ParserExit = ShellCommandArgs()):
    if err.message:
        await matcher.finish(f"参数解析失败：{err.message}")


@cmd_ba_logo.handle()
async def _(matcher: Matcher, arg: Namespace = ShellCommandArgs()):
    try:
        b64_url = await get_logo(arg.text_l, arg.text_r, (not arg.no_transparent))
    except Exception:
        logger.exception("Error when generating image")
        await matcher.finish("遇到错误，请检查后台输出")
    await matcher.finish(MessageSegment.image(b64_url))



import re
from io import BytesIO
from typing import Dict, Tuple

from nonebot.adapters import Message
from nonebot.adapters.onebot.v11 import MessageSegment
from nonebot.matcher import Matcher
from nonebot.params import CommandArg
from pil_utils import Text2Image

from ..command import help_list

usage = "使用指令 `ba帮助` 查询插件功能帮助"
extra = None

FT_REGEX = r"<ft(?P<args>.*?)>(?P<content>.*?)</ft>"


async def t2p(text: str) -> BytesIO:
    t2i = Text2Image.from_bbcode_text(text, fontsize=32)
    img = t2i.to_image("white", (16, 16)).convert("RGB")
    bio = BytesIO()
    img.save(bio, format="jpeg")
    bio.seek(0)
    return bio


async def t2pm(text: str) -> MessageSegment:
    bio = await t2p(text)
    return MessageSegment.image(bio)


def ft_args_to_bbcode(args: str) -> Tuple[str, str]:
    def parse_color(color: str) -> str:
        if color.startswith("(") and color.endswith(")"):
            hex_color = "".join(f"{int(x):02x}" for x in color[1:-1].split(","))
            return f"#{hex_color}"
        return color

    ft_args = dict(x.split("=") for x in args.strip().split())
    bbcode_args: Dict[str, str] = {}

    if "fonts" in ft_args:
        bbcode_args["font"] = ft_args["fonts"]
    if "size" in ft_args:
        bbcode_args["size"] = ft_args["size"]
    if "color" in ft_args:
        bbcode_args["color"] = parse_color(ft_args["color"])

    prefix = ""
    suffix = ""
    for k, v in bbcode_args.items():
        prefix = f"[{k}={v}]{prefix}"
        suffix = f"{suffix}[/{k}]"
    return prefix, suffix


def replace_ft(text: str) -> str:
    def replace(match: re.Match) -> str:
        args = match.group("args")
        prefix, suffix = ft_args_to_bbcode(args)
        content = match.group("content")
        return f"{prefix}{content}{suffix}"

    return re.sub(FT_REGEX, replace, text)


async def help_handle(matcher: Matcher, arg_msg: Message = CommandArg()):
    arg = arg_msg.extract_plain_text().strip()

    if not arg:
        cmd_list = "\n".join(
            (
                f"▸ [b]{k['func']}[/b] "
                f"({k['trigger_method']}：{k['trigger_condition']}) - "
                f"{k['brief_des']}"
            )
            for k in help_list
        )
        msg = (
            f"目前插件支持的功能：\n"
            f"\n"
            f"{cmd_list}\n"
            f"\n"
            f"Tip: 使用指令 `[b]ba帮助 <功能名>[/b]` 查看某功能详细信息"
        )
        await matcher.finish(await t2pm(msg))

    arg_lower = arg.lower()
    func = next(
        (
            x
            for x in help_list
            if (
                (arg_lower in x["func"].lower())
                or (arg_lower in x["trigger_condition"].lower())
                or (arg_lower in x["brief_des"].lower())
            )
        ),
        None,
    )
    if not func:
        await matcher.finish(f"未找到功能 `{arg}`")

    # ft to bbcode
    detail_des = replace_ft(func["detail_des"])

    # 缩进
    detail_des = "    ".join(detail_des.splitlines(keepends=True)).strip()

    msg = (
        f"▸ [b]功能：{func['func']}[/b]\n"
        f"\n"
        f'▹ [b]触发方式：[/b]{func["trigger_method"]}\n'
        f'▹ [b]触发条件：[/b]{func["trigger_condition"]}\n'
        f'▹ [b]简要描述：[/b]{func["brief_des"]}\n'
        f"▹ [b]详细描述：[/b]\n"
        f"    {detail_des}"
    )
    await matcher.finish(await t2pm(msg))



import asyncio
import math
import time
from datetime import datetime, timedelta
from io import BytesIO
from typing import Any, Dict, List, NamedTuple, Optional, cast
from typing_extensions import Unpack

from nonebot_plugin_htmlrender import get_new_page
from PIL import Image, ImageFilter
from PIL.Image import Resampling
from pil_utils import BuildImage, Text2Image, text2image
from playwright.async_api import Page, ViewportSize

from ..config import config
from ..resource import (
    CALENDER_BANNER_PATH,
    GRADIENT_BG_PATH,
    SCHALE_UTIL_CSS_PATH,
    SCHALE_UTIL_JS_PATH,
)
from ..util import (
    AsyncReqKwargs,
    async_req,
    img_invert_rgba,
    parse_time_delta,
    read_image,
    split_list,
)
from ..util import RespType as Rt

PAGE_KWARGS = {
    "is_mobile": True,
    "viewport": ViewportSize(width=767, height=800),
}


async def schale_get(url: str, **kwargs: Unpack[AsyncReqKwargs]) -> Any:
    kwargs["base_urls"] = config.ba_schale_url
    return await async_req(url, **kwargs)


async def schale_get_stu_data(loc: str = "cn") -> List[Dict[str, Any]]:
    return await schale_get(f"data/{loc}/students.min.json")


async def schale_get_config() -> Dict[str, Any]:
    return await schale_get("data/config.min.json")


async def schale_get_localization(loc: str = "cn") -> Dict[str, Any]:
    return cast(dict, await schale_get(f"data/{loc}/localization.min.json"))


async def schale_get_raids(loc: str = "cn") -> Dict[str, Any]:
    return cast(dict, await schale_get(f"data/{loc}/raids.min.json"))


async def schale_get_stu_dict(key: str = "Name") -> dict:
    return {x[key]: x for x in await schale_get_stu_data()}


async def schale_get_stu_info(stu: str) -> bytes:
    async with cast(Page, get_new_page(**PAGE_KWARGS)) as page:
        await page.goto(config.ba_schale_url, wait_until="domcontentloaded")

        await page.goto(
            f"{config.ba_schale_url}?chara={stu}",
            timeout=config.ba_screenshot_timeout * 1000,
            wait_until="networkidle",
        )
        await page.add_style_tag(content=SCHALE_UTIL_CSS_PATH.read_text(encoding="u8"))
        await page.evaluate(SCHALE_UTIL_JS_PATH.read_text(encoding="u8"))
        return await page.screenshot(full_page=True, type="jpeg")


async def schale_calender(server: str) -> BytesIO:
    students, s_config, localization, raids = await asyncio.gather(
        schale_get_stu_dict("Id"),
        schale_get_config(),
        schale_get_localization(),
        schale_get_raids(),
    )
    region_index_map: Dict[str, int] = {
        x["Name"]: i for i, x in enumerate(s_config["Regions"])
    }
    return await schale_get_calender(
        region_index_map[server],
        students,
        s_config,
        localization,
        raids,
    )


class CurrentEventTuple(NamedTuple):
    event: dict
    start: datetime
    end: datetime
    remain: timedelta


def find_current_event(
    ev: List[dict],
    now: Optional[datetime] = None,
) -> Optional[CurrentEventTuple]:
    if not now:
        now = datetime.now()
    for _e in ev:
        _start = datetime.fromtimestamp(_e["start"])
        _end = datetime.fromtimestamp(_e["end"])
        if _start <= now < _end:
            _remain = _end - now
            return CurrentEventTuple(_e, _start, _end, _remain)
    return None


async def schale_get_calender(
    server_index: int,
    students: Dict[str, Dict],
    s_config: dict,
    localization: dict,
    raids: dict,
) -> BytesIO:
    region = s_config["Regions"][server_index]
    now = datetime.now()

    pic_bg = BuildImage.new("RGBA", (1400, 640), (255, 255, 255, 70))

    def format_time(_start: datetime, _end: datetime, _remain: timedelta) -> str:
        dd, hh, mm, ss = parse_time_delta(_remain)
        return (
            f"{_start} ~ {_end} | "
            f"剩余 [b][color=#fc6475]{dd}天 {hh:0>2d}:{mm:0>2d}:{ss:0>2d}[color=#fc6475][/b]"
        )

    async def draw_gacha():
        c_gacha = region["CurrentGacha"]
        if not (r := find_current_event(c_gacha)):
            return None

        pic = pic_bg.copy().draw_text(
            (25, 25, 1375, 150),
            "特选招募",
            weight="bold",
            max_fontsize=80,
        )
        g = r.event
        t = format_time(r.start, r.end, r.remain)
        pic = pic.paste(
            ti := text2image(
                t,
                (255, 255, 255, 0),
                fontsize=45,
                max_width=1350,
                align="center",
            ),
            (int((1400 - ti.width) / 2), 150),
            alpha=True,
        )
        stu = [students[x] for x in g["characters"]]

        async def process_avatar(s):
            return (
                BuildImage.open(
                    BytesIO(
                        await schale_get(
                            f'images/student/collection/{s["Id"]}.webp',
                            resp_type=Rt.BYTES,
                        ),
                    ),
                )
                .resize((300, 340))
                .paste(
                    BuildImage.new("RGBA", (300, 65), (255, 255, 255, 120)),
                    (0, 275),
                    alpha=True,
                )
                .convert("RGB")
                .circle_corner(25)
                .draw_text((0, 275, 300, 340), s["Name"], max_fontsize=50)
            )

        avatars = await asyncio.gather(*[process_avatar(x) for x in stu])
        ava_len = len(avatars)
        x_index = int((1400 - (300 + 25) * ava_len + 25) / 2)

        for p in avatars:
            pic = pic.paste(p, (x_index, 250), alpha=True)
            x_index += p.width + 25

        return pic

    async def draw_event():
        c_event = region["CurrentEvents"]
        if not (r := find_current_event(c_event)):
            return None

        pic = pic_bg.copy().draw_text(
            (25, 25, 1375, 150),
            "当前活动",
            weight="bold",
            max_fontsize=80,
        )

        g = r.event
        t = format_time(r.start, r.end, r.remain)
        pic = pic.paste(
            ti := text2image(
                t,
                (255, 255, 255, 0),
                fontsize=45,
                max_width=1350,
                align="center",
            ),
            (int((1400 - ti.width) / 2), 150),
            alpha=True,
        )
        ev = g["event"]
        ev_name = ""
        if ev >= 10000:
            ev_name = " (复刻)"
            ev %= 10000
        ev_name = localization["EventName"][str(ev)] + ev_name

        ev_bg, ev_img = await asyncio.gather(
            schale_get(
                f"images/campaign/Campaign_Event_{ev}_Normal.png",
                resp_type=Rt.BYTES,
            ),
            schale_get(
                f"images/eventlogo/{ev}_{'Tw' if server_index else 'Jp'}.webp",
                resp_type=Rt.BYTES,
            ),
        )

        ev_bg = (
            BuildImage.open(BytesIO(ev_bg))
            .convert("RGBA")
            .resize_height(340)
            .filter(ImageFilter.GaussianBlur(3))
        )
        ev_bg = (
            ev_bg.paste(
                BuildImage.open(BytesIO(ev_img))
                .convert("RGBA")
                .resize(
                    (ev_bg.width, ev_bg.height - 65),
                    keep_ratio=True,
                    inside=True,
                    bg_color=(255, 255, 255, 0),
                ),
                alpha=True,
            )
            .paste(
                BuildImage.new("RGBA", (ev_bg.width, 65), (255, 255, 255, 120)),
                (0, ev_bg.height - 65),
                alpha=True,
            )
            .convert("RGB")
            .circle_corner(25)
            .draw_text(
                (0, ev_bg.height - 65, ev_bg.width, ev_bg.height),
                ev_name,
                max_fontsize=50,
            )
        )
        return pic.paste(
            ev_bg,
            (int((pic.width - ev_bg.width) / 2), 250),
            alpha=True,
        )

    async def draw_raid():
        if not (r := find_current_event(region["CurrentRaid"])):
            return None

        pic = pic_bg.copy()

        ri = r.event
        t = format_time(r.start, r.end, r.remain)
        pic = pic.paste(
            ti := text2image(
                t,
                (255, 255, 255, 0),
                fontsize=45,
                max_width=1350,
                align="center",
            ),
            (int((1400 - ti.width) / 2), 150),
            alpha=True,
        )

        raid_type = ri["type"]
        time_atk = raid_type == "TimeAttack"
        type_key = "TimeAttack" if time_atk else "Raid"

        raid = {x["Id"]: x for x in raids[type_key]}
        c_ri = raid[ri["raid"]]
        pic = pic.draw_text(
            (25, 25, 1375, 150),
            localization["StageType"][type_key],
            weight="bold",
            max_fontsize=80,
        )

        if time_atk:
            tk_bg = {
                "Shooting": "TimeAttack_SlotBG_02",
                "Defense": "TimeAttack_SlotBG_01",
                "Destruction": "TimeAttack_SlotBG_03",
            }
            bg_url = f'images/timeattack/{tk_bg[c_ri["DungeonType"]]}.png'
            fg_url = f'images/enemy/{c_ri["Icon"]}.webp'
        else:
            bg_url = f'images/raid/Boss_Portrait_{c_ri["PathName"]}_LobbyBG'
            if len(c_ri["Terrain"]) > 1 and ri["terrain"] == c_ri["Terrain"][1]:
                bg_url += f'_{ri["terrain"]}'
            bg_url = f"{bg_url}.png"
            fg_url = f'images/raid/Boss_Portrait_{c_ri["PathName"]}_Lobby.png'
        terrain = c_ri["Terrain"] if time_atk else ri["terrain"]

        color_map = {
            "LightArmor": (167, 12, 25),
            "Explosion": (167, 12, 25),
            "HeavyArmor": (178, 109, 31),
            "Pierce": (178, 109, 31),
            "Unarmed": (33, 111, 156),
            "Mystic": (33, 111, 156),
            "Normal": (115, 115, 115),
        }
        atk_color = color_map[c_ri["BulletType" if time_atk else "BulletTypeInsane"]]
        def_color = color_map[c_ri["ArmorType"]]

        c_bg, c_fg, icon_def, icon_atk, icon_tr = await asyncio.gather(
            *(
                schale_get(bg_url, resp_type=Rt.BYTES),
                schale_get(fg_url, resp_type=Rt.BYTES),
                schale_get("images/ui/Type_Defense_s.png", resp_type=Rt.BYTES),
                schale_get("images/ui/Type_Attack_s.png", resp_type=Rt.BYTES),
                schale_get(f"images/ui/Terrain_{terrain}.png", resp_type=Rt.BYTES),
            ),
        )

        icon_def = (
            BuildImage.new("RGBA", (64, 64), def_color)
            .paste(
                (BuildImage.open(BytesIO(icon_def)).convert("RGBA").resize_height(48)),
                (8, 8),
                alpha=True,
            )
            .circle()
        )
        icon_atk = (
            BuildImage.new("RGBA", (64, 64), atk_color)
            .paste(
                (BuildImage.open(BytesIO(icon_atk)).convert("RGBA").resize_height(48)),
                (8, 8),
                alpha=True,
            )
            .circle()
        )
        icon_tr = (
            BuildImage.new("RGBA", (64, 64), "#ffffff")
            .paste(
                img_invert_rgba(Image.open(BytesIO(icon_tr)).convert("RGBA")),
                (-2, -2),
                alpha=True,
            )
            .circle()
        )

        c_bg = (
            BuildImage.open(BytesIO(c_bg))
            .convert("RGBA")
            .resize_height(340)
            .filter(ImageFilter.GaussianBlur(3))
        )
        c_fg = BuildImage.open(BytesIO(c_fg)).convert("RGBA").resize_height(c_bg.height)
        c_bg = (
            c_bg.paste(
                c_fg,
                (int((c_bg.width - c_fg.width) / 2), 0),
                alpha=True,
            )
            .paste(
                BuildImage.new("RGBA", (c_bg.width, 65), (255, 255, 255, 120)),
                (0, c_bg.height - 65),
                alpha=True,
            )
            .paste(icon_atk, (10, 10), alpha=True)
            .paste(icon_def, (10, 79), alpha=True)
            .paste(icon_tr, (10, 147), alpha=True)
            .convert("RGB")
            .circle_corner(25)
            .draw_text(
                (0, c_bg.height - 65, c_bg.width, c_bg.height),
                (
                    localization["TimeAttackStage"][c_ri["DungeonType"]]
                    if time_atk
                    else (c_ri["Name"])
                ),
                max_fontsize=50,
            )
        )
        return pic.paste(c_bg, (int((pic.width - c_bg.width) / 2), 250), alpha=True)

    async def draw_birth():
        now_t = time.mktime(now.date().timetuple())
        now_w = now.weekday()
        this_week_t = now_t - now_w * 86400
        next_week_t = now_t + (7 - now_w) * 86400
        next_next_week_t = next_week_t + 7 * 86400

        birth_this_week = []
        birth_next_week = []
        for s in [x for x in students.values() if x["IsReleased"][server_index]]:
            try:
                birth = time.mktime(
                    time.strptime(f'{now.year}/{s["BirthDay"]}', "%Y/%m/%d"),
                )
            except ValueError:
                continue
            if this_week_t <= birth < next_week_t:
                birth_this_week.append(s)
            elif next_week_t <= birth <= next_next_week_t:
                birth_next_week.append(s)

        if (not birth_this_week) and (not birth_next_week):
            return None

        img_per_line = 7
        sort_key = lambda x: tuple(  # noqa: E731
            f"{x:0>2}" for x in cast(str, x["BirthDay"]).split("/")
        )
        p_h = 0
        if birth_this_week:
            birth_this_week.sort(key=sort_key)
            p_h += 70 + 220 * math.ceil(len(birth_this_week) / img_per_line)

        if birth_next_week:
            birth_next_week.sort(key=sort_key)
            p_h += 70 + 220 * math.ceil(len(birth_next_week) / img_per_line)
            if birth_this_week:
                p_h += 25

        padding = 25
        min_height = 640
        title_h = 125
        height = max(padding * 3 + title_h + p_h, min_height)
        width = 1400
        pic = BuildImage.new("RGBA", (width, height), (255, 255, 255, 70)).draw_text(
            (25, 25, 1375, 150),
            "学生生日",
            weight="bold",
            max_fontsize=80,
        )
        stu_pics = [
            BuildImage.open(BytesIO(x)).convert("RGBA").resize_height(180).circle()
            for x in await asyncio.gather(
                *(
                    schale_get(
                        f'images/student/icon/{x["Id"]}.webp',
                        resp_type=Rt.BYTES,
                    )
                    for x in birth_this_week + birth_next_week
                ),
            )
        ]

        y_offset = title_h + padding
        if height >= min_height:
            y_offset += (height - (padding // 2) - title_h - p_h) // 2

        def draw_birth_stu(title: str, students: List[dict]):
            nonlocal y_offset

            subtitle = Text2Image.from_text(title, fontsize=45)
            subtitle.draw_on_image(pic.image, ((width - subtitle.width) // 2, y_offset))
            y_offset += 70

            for line in split_list(students, img_per_line):
                x_offset = (width - (190 * len(line))) // 2
                for stu in line:
                    pic.paste(
                        stu_pics.pop(0),
                        (x_offset, y_offset),
                        alpha=True,
                    ).draw_text(
                        (x_offset, y_offset + 180, x_offset + 180, y_offset + 220),
                        stu["BirthDay"],
                    )
                    x_offset += 190
                y_offset += 220

        if birth_this_week:
            draw_birth_stu("本周", birth_this_week)
            y_offset += 25

        if birth_next_week:
            draw_birth_stu("下周", birth_next_week)

        return pic

    img = await asyncio.gather(  # type: ignore
        draw_gacha(),
        draw_event(),
        draw_raid(),
        draw_birth(),
    )
    img: List[BuildImage] = [x for x in img if x]
    if not img:
        img.append(
            pic_bg.copy().draw_text((0, 0, 1400, 640), "没有获取到任何数据", max_fontsize=60),
        )

    bg_w = 1500
    bg_h = 200 + sum([x.height + 50 for x in img])
    bg = (
        BuildImage.new("RGBA", (bg_w, bg_h))
        .paste((await read_image(CALENDER_BANNER_PATH)).resize((1500, 150)))
        .draw_text(
            (50, 0, 1480, 150),
            f"SchaleDB丨活动日程丨{localization['ServerName'][str(server_index)]}",
            max_fontsize=100,
            weight="bold",
            fill="#ffffff",
            halign="left",
        )
        .paste(
            (await read_image(GRADIENT_BG_PATH)).resize(
                (1500, bg_h - 150),
                resample=Resampling.NEAREST,
            ),
            (0, 150),
        )
    )

    h_index = 200
    for im in img:
        bg.paste(im.circle_corner(10), (50, h_index), alpha=True)
        h_index += im.height + 50
    return bg.convert("RGB").save("JPEG")


async def get_fav_li(lvl: int) -> List[dict]:
    return [
        x
        for x in await schale_get_stu_data()
        if (x["MemoryLobby"] and x["MemoryLobby"][0] == lvl)
    ]


async def draw_fav_li(stu_li: List[dict]) -> BytesIO:
    txt_h = 48
    pic_h = 144
    icon_w = 182
    icon_h = pic_h + txt_h
    line_max_icon = 6

    if (li_len := len(stu_li)) <= line_max_icon:
        line = 1
        length = li_len
    else:
        line = math.ceil(li_len / line_max_icon)
        length = line_max_icon

    img = (await read_image(GRADIENT_BG_PATH)).resize(
        (icon_w * length, icon_h * line + 5),
        resample=Resampling.NEAREST,
    )

    async def draw_stu(name_: str, stu_id_: int, line_: int, index_: int):
        left = index_ * icon_w
        top = line_ * icon_h + 5

        ret = await schale_get(
            f"images/student/lobby/{stu_id_}.webp",
            resp_type=Rt.BYTES,
        )
        icon_img = Image.open(BytesIO(ret)).convert("RGBA")
        img.paste(icon_img, (left, top), alpha=True)
        img.draw_text(
            (left, top + pic_h, left + icon_w, top + icon_h),
            name_,
            max_fontsize=25,
            min_fontsize=1,
        )

    task_li = []
    line = 0
    i = 0
    for stu in stu_li:
        if i == line_max_icon:
            i = 0
            line += 1
        task_li.append(draw_stu(stu["Name"], stu["Id"], line, i))
        i += 1
    await asyncio.gather(*task_li)

    return img.convert("RGB").save("JPEG")



import asyncio
import random
from io import BytesIO
from typing import TYPE_CHECKING, List

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Bot, Message, MessageEvent, MessageSegment
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg, EventMessage
from nonebot.typing import T_State
from pil_utils import BuildImage

from ..config import config
from ..data.gamekee import MangaMetadata, get_manga_content, get_manga_list
from ..help import FT_E, FT_S
from ..util import IllegalOperationFinisher, async_req, send_forward_msg, split_list
from ..util import RespType as Rt

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "随机漫画",
        "trigger_method": "指令",
        "trigger_condition": "ba漫画",
        "brief_des": "发送一话官推/同人漫画",
        "detail_des": (
            "从GameKee爬取BA官推/同人漫画并发送\n"
            "可以使用GameKee主页漫画图书馆的漫画名字和链接标题搜索\n"
            "不带参数时，会从所有漫画中随机抽取一话发送\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba漫画{FT_E}\n"
            f"- {FT_S}ba漫画 蔚蓝档案四格漫画{FT_E}\n"
            f"- {FT_S}ba漫画 布噜布噜档案 第一话{FT_E}"
        ),
    },
]


KEY_MANGA_LIST = "manga_list"
KEY_SELECTED_MANGA = "selected_manga"

illegal_finisher = IllegalOperationFinisher("非法操作次数过多，已退出选择")

cmd_random_manga = on_command("ba漫画")


@cmd_random_manga.handle()
async def _(matcher: Matcher, state: T_State, arg_msg: Message = CommandArg()):
    try:
        manga_list = await get_manga_list()
    except Exception:
        logger.exception("获取漫画列表失败")
        await matcher.finish("获取漫画列表失败，请检查后台输出")

    arg = arg_msg.extract_plain_text().strip().split()
    if arg:
        manga_list = [
            x
            for x in manga_list
            if all((kw in x.category or kw in x.name) for kw in arg)
        ]

    if not manga_list:
        await matcher.finish("未找到对应关键词的漫画")

    manga_total = len(manga_list)
    state[KEY_MANGA_LIST] = manga_list
    if manga_total == 1:
        state[KEY_SELECTED_MANGA] = manga_list[0]
    elif not arg:
        index = random.randint(0, manga_total - 1)
        state[KEY_SELECTED_MANGA] = manga_list[index]
    else:
        if manga_total > 5:
            manga_list = manga_list[:5]
            state[KEY_MANGA_LIST] = manga_list
        list_msg = "\n".join(
            f"{i}. 【{x.category}】{x.name}" for i, x in enumerate(manga_list, 1)
        )
        too_much_tip = "\nTip：结果过多，仅显示前五个" if manga_total > 5 else ""
        await matcher.pause(f"找到了多个结果，请发送序号选择，发送 0 退出选择：\n{list_msg}{too_much_tip}")


@cmd_random_manga.handle()
async def _(matcher: Matcher, state: T_State, message: Message = EventMessage()):
    if KEY_SELECTED_MANGA in state:
        return

    arg = message.extract_plain_text().strip()
    if arg == "0":
        await matcher.finish("已退出选择")

    index = int(arg) if arg.isdigit() else None
    manga_list: List[MangaMetadata] = state[KEY_MANGA_LIST]
    if (not index) or (index > len(manga_list)):
        await illegal_finisher()
        await matcher.reject("请输入正确的序号")

    state[KEY_SELECTED_MANGA] = manga_list[index - 1]


@cmd_random_manga.handle()
async def _(
    bot: Bot,
    event: MessageEvent,
    matcher: Matcher,
    state: T_State,
):
    manga: MangaMetadata = state[KEY_SELECTED_MANGA]

    async def get_pic(url):
        p = await async_req(url, resp_type=Rt.BYTES)
        if url.endswith(".webp"):
            p = BuildImage.open(BytesIO(p)).save_png()
        return p

    try:
        content = await get_manga_content(manga.cid)
        pics = await asyncio.gather(*[get_pic(x) for x in content.images])
    except Exception:
        logger.exception(f"获取 CID {manga.cid} 漫画失败")
        await matcher.finish("获取漫画失败，请检查后台输出")

    image_sum = len(pics)
    image_seg = [MessageSegment.image(x) for x in pics]
    if (not config.ba_use_forward_msg) or image_sum <= 2:
        header = (
            f"https://ba.gamekee.com/{manga.cid}.html\n"
            f"【{manga.category}】{content.title}"
        )
        if content.content:
            header = f"{header}\n\n{content.content}"

        chunks = list(split_list(image_seg, 9))
        max_page = len(chunks)
        for i, chunk in enumerate(chunks, 1):
            msg = Message()
            if i == 1:
                msg += header
            msg += chunk
            if max_page > 1:
                msg += f"第 {i} / {max_page} 页（共 {image_sum} P）"
            await matcher.send(msg)
        return

    headers = [Message() + f"【{manga.category}】{content.title}"]
    if content.content:
        headers.append(Message() + content.content)
    images = [Message(x) for x in image_seg]
    footer = Message() + f"https://ba.gamekee.com/{manga.cid}.html"
    await send_forward_msg(bot, event, [*headers, *images, footer])



import shutil
from pathlib import Path

from nonebot import logger

DATA_DIR = Path.cwd() / "data" / "BAWiki"
CACHE_DIR = DATA_DIR / "cache"

for _p in (DATA_DIR, CACHE_DIR):
    if not _p.exists():
        _p.mkdir(parents=True)


RES_DIR = Path(__file__).parent / "res"

RES_GACHA_DIR = RES_DIR / "gacha"
GACHA_BG_PATH = RES_GACHA_DIR / "gacha_bg.webp"
GACHA_BG_OLD_PATH = RES_GACHA_DIR / "gacha_bg_old.webp"
GACHA_CARD_BG_PATH = RES_GACHA_DIR / "gacha_card_bg.png"
GACHA_CARD_MASK_PATH = RES_GACHA_DIR / "gacha_card_mask.png"
GACHA_NEW_PATH = RES_GACHA_DIR / "gacha_new.png"
GACHA_PICKUP_PATH = RES_GACHA_DIR / "gacha_pickup.png"
GACHA_STAR_PATH = RES_GACHA_DIR / "gacha_star.png"
GACHA_STU_ERR_PATH = RES_GACHA_DIR / "gacha_stu_err.png"

RES_GAMEKEE_DIR = RES_DIR / "gamekee"
GAMEKEE_UTIL_JS_PATH = RES_GAMEKEE_DIR / "gamekee_util.js"

RES_GENERAL_DIR = RES_DIR / "general"
CALENDER_BANNER_PATH = RES_GENERAL_DIR / "calender_banner.png"
GRADIENT_BG_PATH = RES_GENERAL_DIR / "gradient.webp"

RES_LOGO_DIR = RES_DIR / "logo"
BA_LOGO_JS_PATH = RES_LOGO_DIR / "ba_logo.js"

RES_SCHALE_DIR = RES_DIR / "schale"
SCHALE_UTIL_JS_PATH = RES_SCHALE_DIR / "schale_util.js"
SCHALE_UTIL_CSS_PATH = RES_SCHALE_DIR / "schale_util.css"

RES_SHITTIM_DIR = RES_DIR / "shittim"
RES_SHITTIM_TEMPLATES_DIR = RES_SHITTIM_DIR / "templates"
RES_SHITTIM_JS_DIR = RES_SHITTIM_DIR / "js"
RES_SHITTIM_CSS_DIR = RES_SHITTIM_DIR / "css"
SHITTIM_UTIL_JS_PATH = RES_SHITTIM_JS_DIR / "shittim_util.js"
SHITTIM_UTIL_CSS_PATH = RES_SHITTIM_CSS_DIR / "shittim_util.css"

EMPTY_HTML_PATH = RES_DIR / "index.html"


_OLD_CACHE_FOLDER = Path.cwd() / "cache"
_OLD_CACHE_PATH = _OLD_CACHE_FOLDER / "BAWiki"
if _OLD_CACHE_PATH.exists():
    logger.warning("Deleting old cache dir...")
    shutil.rmtree(_OLD_CACHE_PATH)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Message, MessageEvent, MessageSegment
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg
from nonebot.typing import T_State

from ..config import config
from ..data.bawiki import db_get_gacha_data, db_get_stu_alias
from ..data.gacha import gacha, get_gacha_cool_down, set_gacha_cool_down
from ..data.schaledb import schale_get_stu_dict
from ..help import FT_E, FT_S
from ..util import recover_alia

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "模拟抽卡",
        "trigger_method": "指令",
        "trigger_condition": "ba抽卡",
        "brief_des": "模拟抽卡",
        "detail_des": (
            "模拟抽卡\n"
            f"可以使用 {FT_S}ba切换卡池{FT_E} 指令来切换卡池\n"
            "可以指定抽卡次数，默认10次\n"
            " \n"
            "可以用这些指令触发：\n"
            f"- {FT_S}ba抽卡{FT_E}\n"
            f"- {FT_S}ba招募{FT_E}\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba抽卡{FT_E}\n"
            f"- {FT_S}ba抽卡 20{FT_E}"
        ),
    },
    {
        "func": "切换卡池",
        "trigger_method": "指令",
        "trigger_condition": "ba切换卡池",
        "brief_des": "设置模拟抽卡的UP池",
        "detail_des": (
            "设置模拟抽卡功能的UP池角色\n"
            "当不带参数时，会展示所有池子以供切换\n"
            f"当参数为 {FT_S}常驻{FT_E} 时，切换到常驻池（没有UP）\n"
            "可以自定义池子UP角色，支持2星与3星角色，参数中学生名称用空格分隔，支持部分学生别名\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba切换卡池{FT_E}\n"
            f"- {FT_S}ba切换卡池 常驻{FT_E}\n"
            f"- {FT_S}ba切换卡池 小桃 小绿{FT_E}"
        ),
    },
]


cmd_change_pool = on_command("ba切换卡池")
cmd_gacha_once = on_command("ba抽卡", aliases={"ba招募"})


@dataclass()
class GachaPool:
    name: str
    pool: List[int]


STATIC_POOL = GachaPool(name="常驻池", pool=[])
gacha_pool_index: Dict[str, GachaPool] = {}


def get_1st_pool(data: dict) -> Optional[GachaPool]:
    if not data:
        return None

    pool_data = data["current_pools"]
    if not pool_data:
        return None

    pool = pool_data[0]
    return GachaPool(name=pool["name"], pool=pool["pool"])


@cmd_change_pool.handle()
async def _(
    matcher: Matcher,
    event: MessageEvent,
    state: T_State,
    cmd_arg: Message = CommandArg(),
):
    arg = cmd_arg.extract_plain_text().strip().lower()
    qq = event.get_user_id()

    if arg:
        if "常驻" in arg:
            current = STATIC_POOL
        else:
            pool = []
            try:
                stu_li = await schale_get_stu_dict()
                stu_alias = await db_get_stu_alias()
            except Exception:
                logger.exception("获取学生列表或别名失败")
                await matcher.finish("获取学生列表或别名失败，请检查后台输出")

            for i in arg.split():
                if not (stu := stu_li.get(recover_alia(i, stu_alias))):
                    await matcher.finish(f"未找到学生 {i}")
                if stu["StarGrade"] == 1:
                    await matcher.finish("不能UP一星角色")
                pool.append(stu)

            current = GachaPool(
                name=f"自定义卡池（{'、'.join([x['Name'] for x in pool])}）",
                pool=[x["Id"] for x in pool],
            )

    else:
        try:
            gacha_data = await db_get_gacha_data()
        except Exception:
            logger.exception("获取抽卡基本数据失败")
            await matcher.finish("获取抽卡基本数据失败，请检查后台输出")

        pool_data = gacha_data["current_pools"]
        first_pool = get_1st_pool(gacha_data)
        if not first_pool:
            await matcher.finish("当前没有可切换的卡池")

        pool_obj = gacha_pool_index.get(qq) or first_pool
        if not pool_obj:
            await matcher.finish("当前没有UP池可供切换")

        if len(pool_data) == 1:
            current = first_pool
        else:
            state["pools"] = pool_data
            pools_str = "\n".join(
                f"{i}. {x['name']}" for i, x in enumerate(pool_data, 1)
            )
            await matcher.pause(f"请选择要切换的卡池：\n{pools_str}\nTip: 发送 0 取消选择")

    if current:
        gacha_pool_index[qq] = current
        await matcher.finish(f"已切换到卡池 {current.name}")


@cmd_change_pool.handle()
async def _(matcher: Matcher, event: MessageEvent, state: T_State):
    index_str = event.get_plaintext().strip()
    if index_str == "0":
        await matcher.finish("已取消选择")

    if not (
        index_str.isdigit() and (1 <= (index := int(index_str)) <= len(state["pools"]))
    ):
        await matcher.reject("请输入有效的序号")

    pools: List[Dict] = state["pools"]
    current = GachaPool(**pools[index - 1])

    gacha_pool_index[event.get_user_id()] = current
    await matcher.finish(f"已切换到卡池 {current.name}")


@cmd_gacha_once.handle()
async def _(
    matcher: Matcher,
    event: MessageEvent,
    cmd_arg: Message = CommandArg(),
):
    session_id = event.get_session_id()

    if cool_down := get_gacha_cool_down(session_id):
        await matcher.finish(f"你先别急，先等 {cool_down} 秒再来抽吧qwq")

    gacha_times = 10
    arg = cmd_arg.extract_plain_text().strip().lower()
    if arg:
        if not arg.isdigit():
            await matcher.finish("请输入有效的整数")
        gacha_times = int(arg)
        if not (
            gacha_times >= 1
            and (config.ba_gacha_max < 1 or gacha_times <= config.ba_gacha_max)
        ):
            await matcher.finish(f"请输入有效的抽卡次数，在1~{config.ba_gacha_max}之间")

    try:
        gacha_data = await db_get_gacha_data()
    except Exception:
        logger.exception("获取抽卡基本数据失败")
        await matcher.finish("获取抽卡基本数据失败，请检查后台输出")

    pool_obj = gacha_pool_index.get(qq := event.get_user_id()) or get_1st_pool(
        gacha_data,
    )
    if not pool_obj:
        await matcher.send("数据源内没有提供当期UP池，已自动切换到常驻池")
        pool_obj = STATIC_POOL
        gacha_pool_index[qq] = STATIC_POOL

    set_gacha_cool_down(session_id)
    try:
        img = await gacha(qq, gacha_times, gacha_data, pool_obj.pool)
    except Exception:
        logger.exception("抽卡错误")
        await matcher.finish("抽卡出错了，请检查后台输出")

    await matcher.finish(
        MessageSegment.at(event.user_id) + f"当前抽取卡池：{pool_obj.name}" + img,
    )



import asyncio
import itertools
import re
import time
import urllib.parse
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from typing import Any, Dict, List, Literal, Optional, cast
from typing_extensions import Unpack

from bs4 import BeautifulSoup, PageElement, ResultSet, Tag
from nonebot import logger
from nonebot_plugin_htmlrender import get_new_page
from PIL import Image
from PIL.Image import Resampling
from pil_utils import BuildImage, text2image
from playwright.async_api import Page

from ..config import config
from ..resource import CALENDER_BANNER_PATH, GAMEKEE_UTIL_JS_PATH, GRADIENT_BG_PATH
from ..util import (
    AsyncReqKwargs,
    async_req,
    i2b,
    parse_time_delta,
    read_image,
    split_pic,
)
from ..util import RespType as Rt


async def game_kee_request(url: str, **kwargs: Unpack[AsyncReqKwargs]) -> Any:
    kwargs["base_urls"] = config.ba_gamekee_url

    headers = kwargs.get("headers") or {}
    headers.update({"Game-Id": "829", "Game-Alias": "ba"})
    kwargs["headers"] = headers

    resp = await async_req(url, **kwargs)
    if resp["code"] != 0:
        raise ValueError(resp["msg"])
    return resp["data"]


async def game_kee_get_calender() -> List[dict]:
    ret = cast(list, await game_kee_request("v1/wiki/index"))

    module = next(
        (x for x in ret if x["module"]["id"] == 12),
        None,
    )
    if not module:
        return []

    li: list = module["list"]

    now = time.time()
    li = [x for x in li if (now < x["end_at"])]

    li.sort(key=lambda x: x["begin_at"] if now < x["begin_at"] else x["end_at"])
    li.sort(key=lambda x: now < x["begin_at"])
    li.sort(key=lambda x: x["importance"], reverse=True)
    return li


async def game_kee_get_stu_li() -> Dict[str, dict]:
    ret = cast(dict, await game_kee_request("v1/wiki/entry"))

    entry_stu = next(
        (x for x in ret["entry_list"] if x["id"] == 23941),
        None,
    )
    if not entry_stu:
        return {}

    entry_stu_all = next(
        (x for x in entry_stu["child"] if x["id"] == 49443),
        None,
    )
    if not entry_stu_all:
        return {}

    return {x["name"]: x for x in entry_stu_all["child"]}


async def game_kee_get_stu_cid_li() -> Dict[str, int]:
    return {x: y["content_id"] for x, y in (await game_kee_get_stu_li()).items()}


def game_kee_page_url(sid: int) -> str:
    return urllib.parse.urljoin(config.ba_gamekee_url, f"{sid}.html")


async def game_kee_get_page(url: str) -> List[BytesIO]:
    async with cast(Page, get_new_page()) as page:
        await page.goto(url, timeout=config.ba_screenshot_timeout * 1000)

        await page.evaluate(GAMEKEE_UTIL_JS_PATH.read_text(encoding="u8"))

        # 太长了
        # 展开折叠的语音
        # folds = await page.query_selector_all("div.fold-table-btn")
        # for i in folds:
        #     with contextlib.suppress(Exception):
        #         await i.click()

        element = await page.query_selector("div.wiki-detail-body")
        assert element
        pic_bytes = await element.screenshot(type="jpeg")

    pic = Image.open(BytesIO(pic_bytes))
    return list(map(i2b, split_pic(pic)))


async def game_kee_calender(
    servers: Optional[List[Literal["Jp", "Global", "Cn"]]] = None,
) -> Optional[List[BytesIO]]:
    ret = await game_kee_get_calender()

    if ret and servers:
        server_name_map = {
            "Jp": "日服",
            "Global": "国际服",
            "Cn": "国服",
        }
        server_names = [server_name_map[x] for x in servers]
        ret = [x for x in ret if x["pub_area"] in server_names]

    if not ret:
        return None

    return await game_kee_get_calender_page(ret)


def split_images(
    images: List[BuildImage],
    max_height: int,
    padding: int,
) -> List[List[BuildImage]]:
    ret = []
    cur = []
    height = 0
    for i in images:
        if height + i.height > max_height:
            ret.append(cur)
            cur = []
            height = 0
        cur.append(i)
        height += i.height + padding
    if cur:
        ret.append(cur)
    return ret


async def game_kee_get_calender_page(
    ret: List[Dict],
    has_pic: bool = True,
) -> List[BytesIO]:
    now = datetime.now()

    async def draw(it: dict) -> BuildImage:
        ev_pic = None
        if has_pic and (url := it.get("picture")):
            try:
                pic_bytes = await async_req(f"https:{url}", resp_type=Rt.BYTES)
                ev_pic = BuildImage.open(BytesIO(pic_bytes))
            except Exception:
                logger.exception("下载日程表图片失败")

        begin = datetime.fromtimestamp(it["begin_at"])
        end = datetime.fromtimestamp(it["end_at"])
        started = begin <= now
        time_remain = (end if started else begin) - now
        dd, hh, mm, ss = parse_time_delta(time_remain)

        # logger.debug(f'{it["title"]} | {started} | {time_remain}')

        title_p = text2image(
            f'[b]{it["title"]}[/b]',
            "#ffffff00",
            max_width=1290,
            fontsize=65,
        )
        time_p = text2image(
            f"{begin} ~ {end}",
            "#ffffff00",
            max_width=1290,
            fontsize=40,
        )
        desc_p = (
            text2image(
                desc.replace("<br>", ""),
                "#ffffff00",
                max_width=1290,
                fontsize=40,
            )
            if (desc := it["description"])
            else None
        )
        remain_p = text2image(
            f"剩余 [color=#fc6475]{dd}[/color] 天 [color=#fc6475]{hh}[/color] 时 "
            f"[color=#fc6475]{mm}[/color] 分 [color=#fc6475]{ss}[/color] 秒"
            f'{"结束" if started else "开始"}',
            "#ffffff00",
            max_width=1290,
            fontsize=50,
        )

        h = (
            100
            + (title_p.height + 25)
            + (time_p.height + 25)
            + (ev_pic.height + 25 if ev_pic else 0)
            + (desc_p.height + 25 if desc_p else 0)
            + remain_p.height
        )
        img = BuildImage.new("RGBA", (1400, h), (255, 255, 255, 70)).draw_rectangle(
            (0, 0, 10, h),
            "#fc6475" if it["importance"] else "#4acf75",
        )

        if not started:
            img.draw_rectangle((1250, 0, 1400, 60), "gray")
            img.draw_text((1250, 0, 1400, 60), "未开始", max_fontsize=50, fill="white")

        ii = 50
        img.paste(title_p, (60, ii), alpha=True)
        ii += title_p.height + 25
        img.paste(time_p, (60, ii), alpha=True)
        ii += time_p.height + 25
        if ev_pic:
            img.paste(ev_pic.resize_width(1290).circle_corner(15), (60, ii), alpha=True)
            ii += ev_pic.height + 25
        if desc_p:
            img.paste(desc_p, (60, ii), alpha=True)
            ii += desc_p.height + 25
        img.paste(remain_p, (60, ii), alpha=True)
        return img

    async def draw_list(li: List[BuildImage], title: str) -> BuildImage:
        bg_w = 1500
        bg_h = 200 + sum([x.height + 50 for x in li])
        bg = (
            BuildImage.new("RGBA", (bg_w, bg_h))
            .paste((await read_image(CALENDER_BANNER_PATH)).resize((1500, 150)))
            .draw_text(
                (50, 0, 1480, 150),
                title,
                max_fontsize=100,
                weight="bold",
                fill="#ffffff",
                halign="left",
            )
            .paste(
                (await read_image(GRADIENT_BG_PATH)).resize(
                    (1500, bg_h - 150),
                    resample=Resampling.NEAREST,
                ),
                (0, 150),
            )
        )

        index = 200
        for p in li:
            bg.paste(p.circle_corner(10), (50, index), alpha=True)
            index += p.height + 50
        return bg

    important_data = []
    common_data = []
    for data in ret:
        (important_data if data["importance"] else common_data).append(data)

    important_pics, common_pics = await asyncio.gather(
        asyncio.gather(*(draw(x) for x in important_data)),
        asyncio.gather(*(draw(x) for x in common_data)),
    )

    max_height = 6000
    if not common_pics:
        pics = [important_pics]
    else:
        chain = itertools.chain(important_pics, common_pics)
        pics: List[List[BuildImage]] = (
            [list(chain)]
            if sum(x.height + 50 for x in chain) <= max_height + 50
            else [important_pics, *split_images(common_pics, max_height, 50)]
        )

    title_prefix = "GameKee丨活动日程"
    if len(pics) == 1:
        images = await asyncio.gather(*(draw_list(pics[0], title_prefix)))
    else:
        if len(pics[-1]) < 3:
            extra = pics.pop()
            pics[-1].extend(extra)
        images = await asyncio.gather(
            *(draw_list(x, f"{title_prefix}丨P{i}") for i, x in enumerate(pics, 1)),
        )

    return [x.convert("RGB").save_jpg() for x in images]


async def game_kee_grab_l2d(cid: int) -> List[str]:
    ret: dict = await game_kee_request(f"v1/content/detail/{cid}")
    content: str = ret["content"]

    soup = BeautifulSoup(content, "lxml")

    l2d_nav_title = soup.find("div", class_="input-wrapper", string="Live2D")
    assert l2d_nav_title
    assert l2d_nav_title.parent
    data_index = l2d_nav_title.parent["data-index"]

    assert l2d_nav_title.parent.parent
    slide_contents = next(
        (x for x in l2d_nav_title.parent.parent.next_siblings if isinstance(x, Tag)),
        None,
    )
    assert slide_contents

    l2d_content = slide_contents.find("div", attrs={"data-index": data_index})
    assert isinstance(l2d_content, Tag)
    images = l2d_content.select(".div-img > img")

    image_urls = [x["data-real"] for x in images]
    return [f"https:{x}" for x in image_urls]


@dataclass()
class GameKeeVoice:
    title: str
    jp: str
    cn: str
    url: str


def parse_voice_elem(elem: Tag) -> GameKeeVoice:
    url: str = cast(str, elem["src"])
    if not url.startswith("http"):
        url = f"https:{url}"

    tr1: Tag = elem.parent.parent.parent.parent  # type: ignore
    tds: ResultSet[Tag] = tr1.find_all("td")
    title = tds[0].text.strip()
    jp = "\n".join(tds[2].stripped_strings)

    tr2 = tr1.next_sibling
    cn = "\n".join(tr2.stripped_strings)  # type: ignore
    return GameKeeVoice(title, jp, cn, url)


def merge_voice_dialogue(voices: List[List[GameKeeVoice]]) -> List[List[GameKeeVoice]]:
    main_voices = voices[0]
    other_voice_entries = voices[1:]

    for voice_entry in other_voice_entries:
        for voice in (x for x in voice_entry if (not x.jp) or (not x.cn)):
            corresponding_voice = next(
                (x for x in main_voices if x.title == voice.title),
                None,
            )
            if not corresponding_voice:
                continue

            if not voice.jp:
                voice.jp = corresponding_voice.jp
            if not voice.cn:
                voice.cn = corresponding_voice.cn

    return voices


async def game_kee_get_voice(cid: int, is_chinese: bool = False) -> List[GameKeeVoice]:
    wiki_html = (
        cast(
            dict,
            await game_kee_request(f"v1/content/detail/{cid}"),
        )
    )["content"]
    bs = BeautifulSoup(wiki_html, "lxml")

    multi_lang_voices = [
        [parse_voice_elem(x) for x in audios]
        for table_fathers in bs.select(".slide-item")
        if (
            # 选择 tables
            (tables := table_fathers.select(".table-overflow > .mould-table"))
            # 检查 tables 中是否有语音，如果有，就取出语音到变量 aus -> audios
            and (
                audios := next(
                    (aus for child in tables if (aus := child.find_all("audio"))),
                    None,
                )
            )
        )
    ]
    if multi_lang_voices:
        return merge_voice_dialogue(multi_lang_voices)[1 if is_chinese else 0]

    # 没有中配
    if is_chinese:
        return []
    return next(
        [parse_voice_elem(x) for x in a]
        for x in bs.select(".mould-table")
        if (a := x.find_all("audio"))
    )


async def get_level_list() -> Dict[str, int]:
    entry = cast(dict, await game_kee_request("v1/wiki/entry"))
    entry_list: List[Dict] = entry["entry_list"]
    guide_entry: List[Dict] = next(x["child"] for x in entry_list if x["id"] == 50284)
    levels = itertools.chain(
        *(x["child"] for x in guide_entry if cast(str, x["name"]).endswith("章")),
    )
    return {
        n.upper(): x["content_id"]
        for x in levels
        if re.match(r"^(H)?(TR|\d+)-(\d+)$", (n := cast(str, x["name"])))
    }


async def extract_content_pic(cid: int) -> List[str]:
    wiki_html = (
        cast(
            dict,
            await game_kee_request(f"v1/content/detail/{cid}"),
        )
    )["content"]
    bs = BeautifulSoup(wiki_html, "lxml")
    img_elem = bs.find_all("img")
    img_urls = cast(List[str], [x["src"] for x in img_elem])
    return [f"https:{x}" if x.startswith("//") else x for x in img_urls]


@dataclass()
class MangaMetadata:
    category: str
    name: str
    cid: int


@dataclass()
class MangaContent:
    title: str
    content: str
    images: List[str]


async def get_manga_list() -> List[MangaMetadata]:
    entry = cast(dict, await game_kee_request("v1/wiki/entry"))
    entry_list: List[Dict] = entry["entry_list"]
    guide_entry: List[Dict] = next(x["child"] for x in entry_list if x["id"] == 51508)

    manga_list: List[MangaMetadata] = []

    current_category = "ぶるーあーかいぶっ！"
    for entry in guide_entry:
        category = entry["name"]
        if category.startswith("【"):
            current_category = category[1 : category.find("】")]
        manga_list.extend(
            MangaMetadata(
                category=current_category,
                name=x["name"],
                cid=x["content_id"],
            )
            for x in entry["child"]
        )

    return manga_list


def tags_to_str(tag: PageElement) -> str:
    def process(elem: PageElement) -> str:
        if c := getattr(elem, "contents", None):
            return "".join([s for x in c if (s := process(x))])
        text = elem if isinstance(elem, str) else elem.text
        if s := text.strip().replace("\u200b", ""):
            return s
        if hasattr(elem, "name") and (elem.name == "img" or elem.name == "br"):  # type: ignore
            return "\n"
        return ""

    text = process(tag).strip()
    if not text:
        return text

    lines = text.splitlines()
    last_line = lines[-1]
    if last_line.strip().endswith(">"):
        lines.pop()
    return "\n".join(lines).strip()


async def get_manga_content(cid: int) -> MangaContent:
    article = cast(dict, await game_kee_request(f"v1/content/detail/{cid}"))
    soup = BeautifulSoup(article["content"], "lxml")

    content = tags_to_str(soup).strip()
    if "汉化：" in content:
        content = content.replace("汉化：", "\n汉化：").replace("\n）", "）")

    return MangaContent(
        title=article["title"],
        content=content,
        images=[
            f"https:{src}"
            for x in soup.find_all("img")
            if (not (src := x["src"]).endswith(".gif")) and "gamekee" in src
        ],
    )



import asyncio
import itertools
from datetime import datetime, timedelta
from enum import Enum, auto
from io import BytesIO
from pathlib import Path
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
    cast,
)
from typing_extensions import ParamSpec, Unpack
from urllib.parse import urljoin
from weakref import WeakSet

import anyio
from async_lru import _LRUCacheWrapper, alru_cache
from httpx import AsyncClient
from nonebot import logger
from nonebot.adapters.onebot.v11 import (
    Bot,
    GroupMessageEvent,
    Message,
    MessageEvent,
    MessageSegment,
)
from nonebot.matcher import current_matcher
from PIL import Image, ImageOps
from pil_utils import BuildImage

from .config import config

T = TypeVar("T")
R = TypeVar("R")
K = TypeVar("K")
V = TypeVar("V")
TC = TypeVar("TC", bound=Callable)
P = ParamSpec("P")
NestedIterable = Iterable[Union[T, Iterable["NestedIterable[T]"]]]
PathType = Union[str, Path, anyio.Path]
SendableType = Union[Message, MessageSegment, str]

KEY_ILLEGAL_COUNT = "_ba_illegal_count"


# region async_req
# 这玩意真的太不优雅了
# 有必要重新写一个 request cache，可以参考 hishel


wrapped_cache_functions: WeakSet["SupportDictCacheWrapper"] = WeakSet()


class SupportDictCacheWrapper(Generic[P, R], _LRUCacheWrapper[R]):  # type: ignore  # ignore final class
    def __init__(
        self,
        fn: Callable[P, Coroutine[Any, Any, R]],
        maxsize: Optional[int] = 128,
        typed: bool = False,
        ttl: Optional[float] = None,
    ) -> None:
        async def wrapped_func(*args, **kwargs):
            new_args, new_kwargs = self._recover_args(args, kwargs)
            return await cast(Callable, fn)(*new_args, **new_kwargs)

        super().__init__(wrapped_func, maxsize, typed, ttl)

    def _process_args(
        self,
        func: Callable[[Any], Any],
        args: Tuple,
        kwargs: Dict,
    ) -> Tuple[Tuple, Dict]:
        new_args = tuple(func(arg) for arg in args)
        new_kwargs = {}
        for k, v in kwargs.items():
            new_kwargs[k] = func(v)
        return new_args, new_kwargs

    def _convert_args(self, args: Tuple, kwargs: Dict) -> Tuple[Tuple, Dict]:
        return self._process_args(
            lambda obj: frozenset(obj.items()) if isinstance(obj, dict) else obj,
            args,
            kwargs,
        )

    def _recover_args(self, args: Tuple, kwargs: Dict) -> Tuple[Tuple, Dict]:
        return self._process_args(
            lambda obj: dict(obj) if isinstance(obj, frozenset) else obj,
            args,
            kwargs,
        )

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        new_args, new_kwargs = self._convert_args(args, kwargs)
        return await super().__call__(*new_args, **new_kwargs)


def wrapped_alru_cache(
    maxsize: Optional[int] = 128,
    typed: bool = False,
    ttl: Optional[int] = None,
):
    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]],
    ) -> SupportDictCacheWrapper[P, R]:
        wrapped = SupportDictCacheWrapper(func, maxsize, typed, ttl)
        wrapped_cache_functions.add(wrapped)
        return wrapped

    return wrapper


class RespType(Enum):
    JSON = auto()
    TEXT = auto()
    BYTES = auto()
    HEADERS = auto()


class AsyncReqKwargs(TypedDict, total=False):
    method: str
    params: Optional[Dict[Any, Any]]
    headers: Optional[Dict[str, Any]]
    content: Union[str, bytes, None]
    data: Optional[Dict[str, Any]]
    json: Optional[Any]
    proxies: Optional[str]

    base_urls: Union[str, List[str]]
    resp_type: RespType
    retries: int
    raise_for_status: bool
    sleep: float


async def base_async_req(*urls: str, **kwargs: Unpack[AsyncReqKwargs]) -> Any:
    if not urls:
        raise ValueError("No URL specified")

    method = kwargs.pop("method", "GET").upper()
    params = kwargs.pop("params", None)
    headers = kwargs.pop("headers", None)
    content = kwargs.pop("content", None)
    data = kwargs.pop("data", None)
    json = kwargs.pop("json", None)
    proxies = kwargs.pop("proxies", config.ba_proxy)

    base_urls = kwargs.pop("base_urls", [])
    resp_type = kwargs.pop("resp_type", RespType.JSON)
    retries = kwargs.pop("retries", config.ba_req_retry)
    raise_for_status = kwargs.pop("raise_for_status", True)
    sleep = kwargs.pop("sleep", 0)

    if base_urls:
        if not isinstance(base_urls, list):
            base_urls = [base_urls]
        urls = tuple(
            urljoin(base_url, url)
            for base_url, url in itertools.product(base_urls, urls)
        )

    async def do_request(current_url: str):
        async with AsyncClient(
            proxies=proxies,
            follow_redirects=True,
            timeout=config.ba_req_timeout,
        ) as cli:
            logger.debug(
                f"{method} `{current_url}`, "
                f"{params=}, {headers=}, {content=}, {data=}, {json=}",
            )
            resp = await cli.request(
                method,
                current_url,
                params=params,
                headers=headers,
                content=content,
                data=data,
                json=json,
            )
            if raise_for_status:
                resp.raise_for_status()

            if sleep:
                await asyncio.sleep(sleep)

            if resp_type == RespType.TEXT:
                return resp.text
            if resp_type == RespType.BYTES:
                return resp.content
            if resp_type == RespType.HEADERS:
                return resp.headers
            return resp.json()  # default RespType.JSON:

    while True:
        url, *rest = urls
        try:
            return await do_request(url)
        except Exception as e:
            e_sfx = f"because error occurred while requesting `{url}`: {e!r}"
            if retries > 0:
                retries -= 1
                logger.error(f"Retrying ({retries} left) {e_sfx}")
            else:
                if not rest:
                    raise ConnectionError("All retries failed") from e
                logger.error(f"Requesting next url `{rest[0]}` {e_sfx}")
                url, *rest = rest
            logger.opt(exception=e).debug("Error Stack")


async_req = wrapped_alru_cache(ttl=config.ba_req_cache_ttl, maxsize=None)(
    base_async_req,
)


def clear_wrapped_alru_cache() -> int:
    cleared = 0
    for wrapped in wrapped_cache_functions:
        size = wrapped.cache_info().currsize
        wrapped.cache_clear()
        cleared += size
    return cleared


# endregion


def format_timestamp(t: int) -> str:
    return datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S")


def recover_alia(origin: str, alia_dict: Dict[str, List[str]]):
    origin = replace_brackets(origin).strip()
    origin_ = origin.lower()

    # 精确匹配
    for k, li in alia_dict.items():
        if origin_ in li or origin_ == k:
            return k

    # 没找到，模糊匹配
    origin_ = origin.replace(" ", "")
    for k, li in alia_dict.items():
        li = [x.replace(" ", "") for x in ([k, *li])]
        for v in li:
            if origin_ in v:
                return k

    return origin


class ParsedTimeDelta(NamedTuple):
    days: int
    hours: int
    minutes: int
    seconds: int


def parse_time_delta(t: timedelta) -> ParsedTimeDelta:
    mm, ss = divmod(t.seconds, 60)
    hh, mm = divmod(mm, 60)
    dd = t.days or 0
    return ParsedTimeDelta(dd, hh, mm, ss)


def img_invert_rgba(im: Image.Image) -> Image.Image:
    # https://stackoverflow.com/questions/2498875/how-to-invert-colors-of-image-with-pil-python-imaging
    r, g, b, a = im.split()
    rgb_image = Image.merge("RGB", (r, g, b))
    inverted_image = ImageOps.invert(rgb_image)
    r2, g2, b2 = inverted_image.split()
    return Image.merge("RGBA", (r2, g2, b2, a))


def replace_brackets(original: str) -> str:
    return original.replace("（", "(").replace("）", "(")


def splice_msg(msgs: Sequence[Union[str, MessageSegment, Message]]) -> Message:
    im = Message()
    for i, v in enumerate(msgs):
        if isinstance(v, str) and (i != 0):
            v = f"\n{v}"
        im += v
    return im


def split_list(lst: Sequence[T], n: int) -> Iterator[Sequence[T]]:
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def split_pic(pic: Image.Image, max_height: int = 4096) -> List[Image.Image]:
    pw, ph = pic.size
    if ph <= max_height:
        return [pic]

    ret = []
    need_merge_last = ph % max_height < max_height // 2
    iter_times = ph // max_height

    now_h = 0
    for i in range(iter_times):
        if i == iter_times - 1 and need_merge_last:
            ret.append(pic.crop((0, now_h, pw, ph)))
            break

        ret.append(pic.crop((0, now_h, pw, now_h + max_height)))
        now_h += max_height

    return ret


def i2b(image: Image.Image, img_format: str = "JPEG") -> BytesIO:
    buf = BytesIO()
    image.save(buf, img_format)
    buf.seek(0)
    return buf


@alru_cache()
async def read_file_cached(path: PathType) -> bytes:
    if not isinstance(path, anyio.Path):
        path = anyio.Path(path)
    return await path.read_bytes()


async def read_image(path: PathType) -> BuildImage:
    content = await read_file_cached(path)
    bio = BytesIO(content)
    return BuildImage.open(bio)


async def send_forward_msg(
    bot: Bot,
    event: MessageEvent,
    messages: Sequence[Union[str, MessageSegment, Message]],
    user_id: Optional[int] = None,
    nickname: Optional[str] = None,
):
    nodes: List[MessageSegment] = [
        MessageSegment.node_custom(
            int(bot.self_id) if user_id is None else user_id,
            "BAWiki" if nickname is None else nickname,
            Message(x),
        )
        for x in messages
    ]
    if isinstance(event, GroupMessageEvent):
        return await bot.send_group_forward_msg(group_id=event.group_id, messages=nodes)
    return await bot.send_private_forward_msg(user_id=event.user_id, messages=nodes)


def camel_case(string: str, upper_first: bool = False) -> str:
    pfx, *rest = string.split("_")
    if upper_first:
        pfx = pfx.capitalize()
    sfx = "".join(x.capitalize() for x in rest)
    return f"{pfx}{sfx}"


class IllegalOperationFinisher:
    def __init__(
        self,
        finish_message: Optional[SendableType] = None,
        limit: int = config.ba_illegal_limit,
    ):
        self.finish_message = finish_message
        self.limit = limit

    async def __call__(
        self,
        finish_message: Union[SendableType, None] = Ellipsis,  # type: ignore
    ):
        if self.limit <= 0:
            return
        matcher = current_matcher.get()
        state = matcher.state

        count = state.get(KEY_ILLEGAL_COUNT, 0) + 1
        if count >= config.ba_illegal_limit:
            await matcher.finish(
                self.finish_message if finish_message is Ellipsis else finish_message,
            )

        state[KEY_ILLEGAL_COUNT] = count



from typing import TYPE_CHECKING

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Message, MessageSegment
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from ..config import config
from ..data.bawiki import recover_stu_alia
from ..data.schaledb import schale_get_stu_dict, schale_get_stu_info
from ..help import FT_E, FT_S

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "学生图鉴",
        "trigger_method": "指令",
        "trigger_condition": "ba学生图鉴",
        "brief_des": "查询学生详情（SchaleDB）",
        "detail_des": (
            "访问对应学生SchaleDB页面并截图，支持部分学生别名\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba学生图鉴 白子{FT_E}\n"
            f"- {FT_S}ba学生图鉴 xcw{FT_E}"
        ),
    },
]


cmd_stu_schale = on_command("ba学生图鉴")


@cmd_stu_schale.handle()
async def _(matcher: Matcher, cmd_arg: Message = CommandArg()):
    arg = cmd_arg.extract_plain_text().strip()
    if not arg:
        await matcher.finish("请提供学生名称")

    try:
        ret = await schale_get_stu_dict()
    except Exception:
        logger.exception("获取学生列表出错")
        await matcher.finish("获取学生列表表出错，请检查后台输出")

    if not ret:
        await matcher.finish("没有获取到学生列表数据")

    if not (data := ret.get(await recover_stu_alia(arg))):
        await matcher.finish("未找到该学生")

    stu_name = data["PathName"]
    await matcher.send(f"请稍等，正在截取SchaleDB页面～\n{config.ba_schale_url}?chara={stu_name}")

    try:
        img = MessageSegment.image(await schale_get_stu_info(stu_name))
    except Exception:
        logger.exception(f"截取schale db页面出错 chara={stu_name}")
        await matcher.finish("截取页面出错，请检查后台输出")

    await matcher.finish(img)



from typing import TYPE_CHECKING, List, NoReturn

from nonebot import logger, on_command, on_shell_command
from nonebot.adapters.onebot.v11 import Message, MessageEvent, MessageSegment
from nonebot.exception import ParserExit
from nonebot.matcher import Matcher
from nonebot.params import ArgPlainText, CommandArg, ShellCommandArgs
from nonebot.permission import SUPERUSER
from nonebot.rule import ArgumentParser, Namespace
from nonebot.typing import T_State

from ..config import config
from ..data.arona import ImageModel, get_image, search, search_exact, set_alias
from ..help import FT_E, FT_S
from ..util import IllegalOperationFinisher

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "Arona数据源攻略",
        "trigger_method": "指令",
        "trigger_condition": "arona",
        "brief_des": "从 Arona Bot 数据源中搜索攻略图",
        "detail_des": (
            "从 Arona Bot 数据源中搜索攻略图，支持模糊搜索\n"
            "感谢 diyigemt 佬的开放 API 数据源\n"
            " \n"
            f"可以使用 {FT_S}arona设置别名{FT_E} 指令来为某关键词设置别名\n"
            f"使用方式：{FT_S}arona设置别名 原名 别名1 别名2 ...{FT_E}\n"
            "\n"
            f"可以使用 {FT_S}arona删除别名{FT_E} 指令来删除已设置的关键词的别名\n"
            f"使用方式：{FT_S}arona删除别名 别名1 别名2 ...{FT_E}\n"
            " \n"
            "可以搜索的内容：\n"
            f"- {FT_S}学生攻略图{FT_E}（星野，白子 等）\n"
            f"- {FT_S}主线地图{FT_E}（1-1，H1-1 等）\n"
            f"- {FT_S}杂图{FT_E}（使用 {FT_S}arona 杂图{FT_E} 可以获取图片列表）\n"
            " \n"
            "可以用这些指令触发：\n"
            f"- {FT_S}arona{FT_E}\n"
            f"- {FT_S}蓝色恶魔{FT_E}\n"
            f"- {FT_S}Arona{FT_E}\n"
            f"- {FT_S}ARONA{FT_E}\n"
            f"- {FT_S}阿罗娜{FT_E}\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}arona{FT_E}（会向你提问需要搜索什么）\n"
            f"- {FT_S}arona 国际服未来视{FT_E}（精确搜索）\n"
            f"- {FT_S}arona 国际服{FT_E}（模糊搜索）"
        ),
    },
]


illegal_finisher = IllegalOperationFinisher("坏老师，一直逗我，不理你了，哼！")

ARONA_PREFIXES = ["arona", "蓝色恶魔", "Arona", "ARONA", "阿罗娜"]
cmd_arona = on_command(ARONA_PREFIXES[0], aliases=set(ARONA_PREFIXES[1:]), priority=2)


async def send_image(matcher: Matcher, img: ImageModel) -> NoReturn:
    try:
        res = await get_image(img.path, img.hash)
    except Exception:
        logger.exception("Arona数据源图片获取失败")
        await matcher.finish("呜呜，阿罗娜在获取图片的时候遇到了点问题 QAQ")

    await matcher.finish(MessageSegment.image(res))


@cmd_arona.handle()
async def _(matcher: Matcher, arg: Message = CommandArg()):
    if arg.extract_plain_text().strip():
        matcher.set_arg("param", arg)


@cmd_arona.got("param", prompt="老师，请发送想要搜索的内容~")
async def _(matcher: Matcher, state: T_State, param: str = ArgPlainText()):
    param = param.strip()
    if not param:
        await illegal_finisher()
        await matcher.reject("老师真是的，快给我发送你想要搜索的内容吧！")

    try:
        res = await search(param)
    except Exception:
        logger.exception("Arona数据源搜索失败")
        await matcher.finish("呜呜，阿罗娜在搜索结果的时候遇到了点问题 QAQ")

    if not res:
        await matcher.finish("抱歉老师，阿罗娜没有找到相关结果……")

    if len(res) == 1:
        await send_image(matcher, res[0])

    state["res"] = res
    list_txt = "\n".join(f"{i}. {r.name}" for i, r in enumerate(res, 1))
    await matcher.pause(
        f"老师！阿罗娜帮您找到了多个可能的结果，请发送序号来选择吧~\n{list_txt}\nTip：发送 0 取消选择",
    )


@cmd_arona.handle()
async def _(event: MessageEvent, matcher: Matcher, state: T_State):
    index_str = event.get_plaintext().strip()
    res: List[ImageModel] = state["res"]

    if index_str == "0":
        await matcher.finish("OK，阿罗娜已经取消老师的选择了~")

    if not index_str.isdigit():
        await illegal_finisher()
        await matcher.reject("不要再逗我了，老师！快发送你要选择的序号吧 quq")

    index = int(index_str)
    if not (0 <= index <= len(res)):
        await illegal_finisher()
        await matcher.reject("抱歉，阿罗娜找不到老师发送的序号哦，请老师重新发送一下吧")

    param = res[index - 1].name
    try:
        final_res = await search(param)
        assert final_res
    except Exception:
        logger.exception("Arona数据源搜索失败")
        await matcher.finish("呜呜，阿罗娜在搜索结果的时候遇到了点问题 QAQ")

    await send_image(matcher, final_res[0])


ARONA_SET_ALIAS_COMMANDS = [f"{p}设置别名" for p in ARONA_PREFIXES]
ARONA_DEL_ALIAS_COMMANDS = [f"{p}删除别名" for p in ARONA_PREFIXES]

cmd_arona_set_alias_parser = ArgumentParser(ARONA_SET_ALIAS_COMMANDS[0])
cmd_arona_set_alias_parser.add_argument("name", help="原名")
cmd_arona_set_alias_parser.add_argument("aliases", nargs="+", help="别名，可以提供多个")
cmd_arona_set_alias = on_shell_command(
    ARONA_SET_ALIAS_COMMANDS[0],
    aliases=set(ARONA_SET_ALIAS_COMMANDS[1:]),
    parser=cmd_arona_set_alias_parser,
    permission=SUPERUSER if config.ba_arona_set_alias_only_su else None,
)

cmd_aro_del_alias_parser = ArgumentParser(ARONA_DEL_ALIAS_COMMANDS[0])
cmd_aro_del_alias_parser.add_argument("aliases", nargs="+", help="别名，可以提供多个")
cmd_arona_del_alias = on_shell_command(
    ARONA_DEL_ALIAS_COMMANDS[0],
    aliases=set(ARONA_DEL_ALIAS_COMMANDS[1:]),
    parser=cmd_aro_del_alias_parser,
    permission=SUPERUSER if config.ba_arona_set_alias_only_su else None,
)


@cmd_arona_set_alias.handle()
@cmd_arona_del_alias.handle()
async def _(matcher: Matcher, foo: ParserExit = ShellCommandArgs()):
    await matcher.finish(foo.message)


@cmd_arona_set_alias.handle()
async def _(matcher: Matcher, args: Namespace = ShellCommandArgs()):
    try:
        assert isinstance(args.name, str)
        assert all(isinstance(a, str) for a in args.aliases)
    except AssertionError:
        await matcher.finish("请老师发送纯文本消息的说")

    name: str = args.name.strip()
    aliases: List[str] = [a.strip().lower() for a in args.aliases]

    try:
        resp = await search_exact(name)
    except Exception:
        logger.exception(f"Arona数据源搜索失败 {args.name}")
        await matcher.finish("抱歉，阿罗娜在尝试查找原名是否存在时遇到了一点小问题……")

    if (not resp) or (len(resp) > 1):
        await matcher.finish("啊咧？阿罗娜找不到老师提供的原名，请老师检查一下您提供的名称是否正确")

    ret_dict = set_alias(name, aliases)
    message = "\n".join(
        (
            "好的，阿罗娜已经成功帮你操作了以下别名~",
            *(
                (f"成功将别名 {k} 指向的原名从 {v} 更改为 {name}" if v else f"成功设置 {k} 为 {name} 的别名")
                for k, v in ret_dict.items()
            ),
        ),
    )
    await matcher.finish(message)


@cmd_arona_del_alias.handle()
async def _(matcher: Matcher, args: Namespace = ShellCommandArgs()):
    try:
        assert all(isinstance(a, str) for a in args.aliases)
    except AssertionError:
        await matcher.finish("请老师发送纯文本消息的说")

    aliases: List[str] = [a.strip().lower() for a in args.aliases]

    ret_dict = set_alias(None, aliases)
    message = "\n".join(
        (
            "好的，阿罗娜已经成功帮你操作了以下别名~",
            *(
                (f"成功删除指向原名 {v} 的别名 {k} " if v else f"已设置的别名中未找到 {k}")
                for k, v in ret_dict.items()
            ),
        ),
    )
    await matcher.finish(message)



import asyncio
from typing import TYPE_CHECKING, Awaitable, Callable, Dict, Optional, Tuple

from nonebot import logger, on_command
from nonebot.adapters.onebot.v11 import MessageSegment
from nonebot.adapters.onebot.v11.message import Message
from nonebot.matcher import Matcher
from nonebot.params import CommandArg
from nonebot.typing import T_State

from ..config import config
from ..data.shittim_chest import (
    RAID_ANALYSIS_URL,
    RANK_DATA_TYPE_NAME_MAP,
    SERVER_NAME_MAP,
    RankDataType,
    RankRecord,
    Season,
    ServerType,
    async_iter_all,
    get_alice_friends,
    get_diligent_achievers,
    get_participation_chart_data,
    get_raid_chart_data,
    get_rank_list,
    get_rank_list_by_last_rank,
    get_rank_list_top,
    get_season_list,
    render_raid_analysis,
    render_raid_rank,
    render_rank_detail,
)
from ..help import FT_E, FT_S

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "国服总力数据",
        "trigger_method": "指令",
        "trigger_condition": "ba帮助 国服总力",
        "brief_des": "从什亭之匣查询国服总力相关数据",
        "detail_des": (
            "查询国服总力相关数据\n"
            f"数据来源：什亭之匣（{config.ba_shittim_url}）\n"
            " \n"
            "下面指令中的可选参数介绍，每个参数间需以空格分隔：\n"
            f"- {FT_S}服务器名称{FT_E} 可选值 “官服” 或 “B服”，默认为官服；\n"
            f"- {FT_S}期数序号{FT_E} 可以使用指令 {FT_S}ba总力列表{FT_E} 查询，默认为最新一期；\n"
            " \n"
            "使用以下指令查询总力档线：（可选参数：服务器名称、期数序号）\n"
            f"- {FT_S}ba总力档线{FT_E}\n"
            f"- {FT_S}ba档线{FT_E}\n"
            " \n"
            "使用以下指令查询总力统计概览：（无参数）\n"
            f"- {FT_S}ba总力统计{FT_E}\n"
            " \n"
            "使用以下指令查询往期总力第 1 名：（可选参数：服务器名称）\n"
            f"- {FT_S}ba小心卷狗{FT_E}\n"
            " \n"
            "使用以下指令查询往期总力第 20001 名：（可选参数：服务器名称）\n"
            f"- {FT_S}ba爱丽丝的伙伴{FT_E}\n"
            f"- {FT_S}ba爱丽丝伙伴{FT_E}"
        ),
    },
]


def parse_args(raw_arg: str) -> Tuple[ServerType, Optional[int]]:
    args = raw_arg.strip().split()

    server = ServerType.Official
    season: Optional[int] = None

    for arg in args:
        if server_id := next(
            (k for k, v in SERVER_NAME_MAP.items() if v.lower() == arg.lower()),
            None,
        ):
            server = ServerType(server_id)
        elif arg.isdigit():
            season = int(arg)
        else:
            raise ValueError(f"参数 `{arg}` 无效")

    return server, season


async def get_season_by_index(season_index: Optional[int]) -> Season:
    seasons = await get_season_list()
    return (
        seasons[0]
        if season_index is None
        else next(x for x in seasons if x.season == season_index)
    )


cmd_raid_list = on_command("ba总力列表")


@cmd_raid_list.handle()
async def _(matcher: Matcher):
    try:
        seasons = await get_season_list()
    except Exception:
        logger.exception("Error when getting rank data")
        await matcher.finish("获取数据时出错，请检查后台输出")

    if not seasons:
        await matcher.finish("暂无数据")

    # seasons.sort(key=lambda season: season.season, reverse=True)
    await matcher.finish(
        "\n".join(
            f"第 {season.season} 期 - {season.season_map.value} {season.boss}"
            for season in seasons
        ),
    )


cmd_raid_score = on_command(
    "ba总力档线",
    aliases={"ba档线"},
    state={"data_type": RankDataType.Score},
)


@cmd_raid_score.handle()
async def _(matcher: Matcher, state: T_State, arg_msg: Message = CommandArg()):
    try:
        server, season_index = parse_args(arg_msg.extract_plain_text())
    except ValueError as e:
        await matcher.finish(str(e))

    data_type: RankDataType = state["data_type"]

    server_name = SERVER_NAME_MAP[server.value]
    data_type_name = RANK_DATA_TYPE_NAME_MAP[data_type.value]

    try:
        season = await get_season_by_index(season_index)
        season_index = season.season
        (
            rank_list_top,
            rank_list_by_last_rank,
            rank_list,
            raid_chart,
            participation_chart,
        ) = await asyncio.gather(
            get_rank_list_top(server, season_index),
            get_rank_list_by_last_rank(server, season_index),
            async_iter_all(get_rank_list(server, data_type, season_index)),
            get_raid_chart_data(server, season_index),
            get_participation_chart_data(server, season_index),
        )
    except StopIteration:
        await matcher.finish("期数不存在")
    except Exception:
        logger.exception("Error when getting rank data")
        await matcher.finish("获取数据时出错，请检查后台输出")

    if not rank_list:
        await matcher.finish(f"暂无{data_type_name}数据")

    try:
        img = await render_raid_rank(
            server_name,
            data_type_name,
            season,
            rank_list_top,
            rank_list_by_last_rank,
            rank_list,
            raid_chart,
            participation_chart,
        )
    except Exception:
        logger.exception("Error when rendering image")
        await matcher.finish("渲染图片时出错，请检查后台输出")

    await matcher.finish(MessageSegment.image(img))


cmd_raid_analysis = on_command("ba总力统计")


@cmd_raid_analysis.handle()
async def _(matcher: Matcher):
    try:
        img = await render_raid_analysis()
    except Exception:
        logger.exception("Error when rendering image")
        await matcher.finish("渲染图片时出错，请检查后台输出")

    await matcher.finish(MessageSegment.image(img) + f"详情请访问 {RAID_ANALYSIS_URL}")


cmd_alice_friends = on_command(
    "ba爱丽丝的伙伴",
    aliases={"ba爱丽丝伙伴"},
    state={"func": get_alice_friends, "title": "爱丽丝的伙伴"},
)
cmd_diligent_achievers = on_command(
    "ba小心卷狗",
    state={"func": get_diligent_achievers, "title": "小心卷狗"},
)


@cmd_alice_friends.handle()
@cmd_diligent_achievers.handle()
async def _(matcher: Matcher, state: T_State, arg_msg: Message = CommandArg()):
    try:
        server, _ = parse_args(arg_msg.extract_plain_text())
    except ValueError as e:
        await matcher.finish(str(e))

    title: str = f'{state["title"]}（{SERVER_NAME_MAP[server.value]}）'
    func: Callable[[ServerType], Awaitable[Dict[int, RankRecord]]] = state["func"]

    try:
        season_list, rank_list = await asyncio.gather(get_season_list(), func(server))
    except Exception:
        logger.exception("Error when getting rank data")
        await matcher.finish("获取数据时出错，请检查后台输出")

    try:
        img = await render_rank_detail(title, season_list, rank_list)
    except Exception:
        logger.exception("Error when rendering image")
        await matcher.finish("渲染图片时出错，请检查后台输出")

    await matcher.finish(MessageSegment.image(img))



from typing import TYPE_CHECKING

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Bot, Message, MessageEvent, MessageSegment
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from ..config import config
from ..data.bawiki import recover_stu_alia
from ..data.gamekee import (
    game_kee_get_page,
    game_kee_get_stu_cid_li,
    game_kee_page_url,
)
from ..help import FT_E, FT_S
from ..util import send_forward_msg

if TYPE_CHECKING:
    from . import HelpList

help_list: "HelpList" = [
    {
        "func": "学生Wiki",
        "trigger_method": "指令",
        "trigger_condition": "ba学生wiki",
        "brief_des": "查询学生详情（GameKee）",
        "detail_des": (
            "访问对应学生GameKee Wiki页面并截图，支持部分学生别名\n"
            " \n"
            "指令示例：\n"
            f"- {FT_S}ba学生wiki 白子{FT_E}\n"
            f"- {FT_S}ba学生wiki xcw{FT_E}"
        ),
    },
]


cmd_stu_wiki = on_command("ba学生wiki", aliases={"ba学生Wiki", "ba学生WIKI"})


@cmd_stu_wiki.handle()
async def _(
    bot: Bot,
    event: MessageEvent,
    matcher: Matcher,
    cmd_arg: Message = CommandArg(),
):
    arg = cmd_arg.extract_plain_text().strip()
    if not arg:
        await matcher.finish("请提供学生名称")

    try:
        ret = await game_kee_get_stu_cid_li()
    except Exception:
        logger.exception("获取学生列表出错")
        await matcher.finish("获取学生列表出错，请检查后台输出")

    if not ret:
        await matcher.finish("没有获取到学生列表数据")

    if not (sid := ret.get(await recover_stu_alia(arg, game_kee=True))):
        await matcher.finish("未找到该学生")

    url = game_kee_page_url(sid)
    await matcher.send(f"请稍等，正在截取Wiki页面……\n{url}")

    try:
        images = await game_kee_get_page(url)
    except Exception:
        logger.exception(f"截取wiki页面出错 {url}")
        await matcher.finish("截取页面出错，请检查后台输出")

    img_seg = [MessageSegment.image(x) for x in images]
    if config.ba_use_forward_msg:
        await send_forward_msg(bot, event, img_seg)
    else:
        await matcher.finish(Message(img_seg))



import hashlib
import json
import shutil
from typing import Dict, List, Optional

import anyio
from pydantic import BaseModel

from ..config import config
from ..resource import CACHE_DIR, DATA_DIR
from ..util import RespType as Rt
from ..util import async_req

ARONA_CACHE_DIR = CACHE_DIR / "arona"
if config.ba_auto_clear_cache_path and ARONA_CACHE_DIR.exists():
    shutil.rmtree(ARONA_CACHE_DIR)
if not ARONA_CACHE_DIR.exists():
    ARONA_CACHE_DIR.mkdir(parents=True)

ARONA_ALIAS_PATH = DATA_DIR / "arona_alias.json"  # {"alias": "name"}
if not ARONA_ALIAS_PATH.exists():
    ARONA_ALIAS_PATH.write_text("{}", encoding="u8")


class ImageModel(BaseModel):
    name: str
    path: str
    hash: str  # noqa: A003
    type: int  # noqa: A003


class ImageAPIResult(BaseModel):
    status: int
    data: Optional[List[ImageModel]]
    message: str


async def get_image(path: str, hash_str: Optional[str] = None) -> bytes:
    if not path.startswith("/"):
        raise ValueError("path must start with /")

    if hash_str:
        file_path = anyio.Path(ARONA_CACHE_DIR / hash_str)
        if await file_path.exists():
            return await file_path.read_bytes()

    content = await async_req(
        f"image{path}",
        base_urls=config.ba_arona_cdn_url,
        resp_type=Rt.BYTES,
    )

    if not hash_str:
        hash_str = hashlib.md5(content).hexdigest()

    file_path = anyio.Path(ARONA_CACHE_DIR / hash_str)
    await file_path.write_bytes(content)

    return content


async def search_exact(name: str) -> Optional[List[ImageModel]]:
    resp: dict = await async_req(
        "api/v1/image",
        base_urls=config.ba_arona_api_url,
        params={"name": name},
    )
    return ImageAPIResult(**resp).data


async def search(name: str) -> Optional[List[ImageModel]]:
    alias_dict = json.loads(ARONA_ALIAS_PATH.read_text(encoding="u8"))
    recovered = alias_dict[lowered] if (lowered := name.lower()) in alias_dict else None
    if recovered and (resp := await search_exact(recovered)):
        return resp
    return await search_exact(name)


def set_alias(name: Optional[str], alias: List[str]) -> Dict[str, Optional[str]]:
    data = json.loads(ARONA_ALIAS_PATH.read_text(encoding="u8"))
    original_dict: Dict[str, Optional[str]] = {}

    for a in alias:
        original_dict[a] = data.get(a)
        if name is None:
            data.pop(a, None)
        else:
            data[a] = name

    ARONA_ALIAS_PATH.write_text(json.dumps(data, ensure_ascii=False), encoding="u8")
    return original_dict
